# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Soccer observables modules."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

from dm_control.composer.observation import observable as base_observable
from dm_control.locomotion.soccer import team as team_lib
import numpy as np
import six
from six.moves import zip


@six.add_metaclass(abc.ABCMeta)
class ObservablesAdder(object):
  """A callable that adds a set of per-player observables for a task."""

  @abc.abstractmethod
  def __call__(self, task, player):
    """Adds observables to a player for the given task.

    Args:
      task: A `soccer.Task` instance.
      player: A `Walker` instance to which observables will be added.
    """


class MultiObservablesAdder(ObservablesAdder):
  """Applies multiple `ObservablesAdder`s to a soccer task and player."""

  def __init__(self, observables):
    """Initializes a `MultiObservablesAdder` instance.

    Args:
      observables: A list of `ObservablesAdder` instances.
    """
    self._observables = observables

  def __call__(self, task, player):
    """Adds observables to a player for the given task.

    Args:
      task: A `soccer.Task` instance.
      player: A `Walker` instance to which observables will be added.
    """
    for observable in self._observables:
      observable(task, player)


class CoreObservablesAdder(ObservablesAdder):
  """Core set of per player observables."""

  def __call__(self, task, player):
    """Adds observables to a player for the given task.

    Args:
      task: A `soccer.Task` instance.
      player: A `Walker` instance to which observables will be added.
    """
    # Enable proprioceptive observables.
    self._add_player_proprio_observables(player)

    # Add egocentric observations of soccer ball.
    self._add_player_observables_on_ball(player, task.ball)

    # Add egocentric observations of others.
    teammate_id = 0
    opponent_id = 0
    for other in task.players:
      if other is player:
        continue
      # Infer team prefix for `other` conditioned on `player.team`.
      if player.team != other.team:
        prefix = 'opponent_{}'.format(opponent_id)
        opponent_id += 1
      else:
        prefix = 'teammate_{}'.format(teammate_id)
        teammate_id += 1

      self._add_player_observables_on_other(player, other, prefix)

    self._add_player_arena_observables(player, task.arena)

    # Add per player game statistics.
    self._add_player_stats_observables(task, player)

  def _add_player_observables_on_other(self, player, other, prefix):
    """Add observables of another player in this player's egocentric frame.

    Args:
      player: A `Walker` instance, the player we are adding observables to.
      other: A `Walker` instance corresponding to a different player.
      prefix: A string specifying a prefix to apply to the names of observables
        belonging to `player`.
    """
    if player is other:
      raise ValueError('Cannot add egocentric observables of player on itself.')
    # Origin callable in xpos, xvel for `player`.
    xpos_xyz_callable = lambda p: p.bind(player.walker.root_body).xpos
    xvel_xyz_callable = lambda p: p.bind(player.walker.root_body).cvel[3:]
    # Egocentric observation of other's position, orientation and
    # linear velocities.
    def _cvel_observation(physics, other=other):
      # Velocitmeter reads in local frame but we need world frame observable
      # for egocentric transformation.
      return physics.bind(other.walker.root_body).cvel[3:]

    def _egocentric_end_effectors_xpos(physics, other=other):
      origin_xpos = xpos_xyz_callable(physics)
      egocentric_end_effectors_xpos = []
      for end_effector_body in other.walker.end_effectors:
        xpos = physics.bind(end_effector_body).xpos
        delta = xpos - origin_xpos
        ego_xpos = player.walker.transform_vec_to_egocentric_frame(
            physics, delta)
        egocentric_end_effectors_xpos.append(ego_xpos)
      return np.concatenate(egocentric_end_effectors_xpos)

    player.walker.observables.add_egocentric_vector(
        '{}_ego_linear_velocity'.format(prefix),
        base_observable.Generic(_cvel_observation),
        origin_callable=xvel_xyz_callable)
    player.walker.observables.add_egocentric_vector(
        '{}_ego_position'.format(prefix),
        other.walker.observables.position,
        origin_callable=xpos_xyz_callable)
    player.walker.observables.add_egocentric_xmat(
        '{}_ego_orientation'.format(prefix),
        other.walker.observables.orientation)

    # Adds end effectors of the other agents in the player's egocentric frame.
    player.walker.observables.add_observable(
        '{}_ego_end_effectors_pos'.format(prefix),
        base_observable.Generic(_egocentric_end_effectors_xpos))

    # Adds end effectors of the other agents in the other's egocentric frame.
    # A is seeing B's hand extended to B's right.
    player.walker.observables.add_observable(
        '{}_end_effectors_pos'.format(prefix),
        other.walker.observables.end_effectors_pos)

  def _add_player_observables_on_ball(self, player, ball):
    """Add observables of the soccer ball in this player's egocentric frame.

    Args:
      player: A `Walker` instance, the player we are adding observations for.
      ball: A `SoccerBall` instance.
    """
    # Origin callables for egocentric transformations.
    xpos_xyz_callable = lambda p: p.bind(player.walker.root_body).xpos
    xvel_xyz_callable = lambda p: p.bind(player.walker.root_body).cvel[3:]

    # Add egocentric ball observations.
    player.walker.observables.add_egocentric_vector(
        'ball_ego_angular_velocity', ball.observables.angular_velocity)
    player.walker.observables.add_egocentric_vector(
        'ball_ego_position',
        ball.observables.position,
        origin_callable=xpos_xyz_callable)
    player.walker.observables.add_egocentric_vector(
        'ball_ego_linear_velocity',
        ball.observables.linear_velocity,
        origin_callable=xvel_xyz_callable)

  def _add_player_proprio_observables(self, player):
    """Add proprioceptive observables to the given player.

    Args:
      player: A `Walker` instance, the player we are adding observations for.
    """
    for observable in (player.walker.observables.proprioception +
                       player.walker.observables.kinematic_sensors):
      observable.enabled = True

    # Also enable previous action observable as part of proprioception.
    player.walker.observables.prev_action.enabled = True

  def _add_player_arena_observables(self, player, arena):
    """Add observables of the arena.

    Args:
      player: A `Walker` instance to which observables will be added.
      arena: A `Pitch` instance.
    """
    # Enable egocentric view of position detectors (goal, field).
    # Corners named according to walker *facing towards opponent goal*.
    clockwise_names = [
        'team_goal_back_right',
        'team_goal_mid',
        'team_goal_front_left',
        'field_front_left',
        'opponent_goal_back_left',
        'opponent_goal_mid',
        'opponent_goal_front_right',
        'field_back_right',
    ]
    clockwise_features = [
        lambda _: arena.home_goal.lower[:2],
        lambda _: arena.home_goal.mid,
        lambda _: arena.home_goal.upper[:2],
        lambda _: arena.field.upper,
        lambda _: arena.away_goal.upper[:2],
        lambda _: arena.away_goal.mid,
        lambda _: arena.away_goal.lower[:2],
        lambda _: arena.field.lower,
    ]
    xpos_xyz_callable = lambda p: p.bind(player.walker.root_body).xpos
    xpos_xy_callable = lambda p: p.bind(player.walker.root_body).xpos[:2]
    # A list of egocentric reference origin for each one of clockwise_features.
    clockwise_origins = [
        xpos_xy_callable,
        xpos_xyz_callable,
        xpos_xy_callable,
        xpos_xy_callable,
        xpos_xy_callable,
        xpos_xyz_callable,
        xpos_xy_callable,
        xpos_xy_callable,
    ]
    if player.team != team_lib.Team.HOME:
      half = len(clockwise_features) // 2
      clockwise_features = clockwise_features[half:] + clockwise_features[:half]
      clockwise_origins = clockwise_origins[half:] + clockwise_origins[:half]

    for name, feature, origin in zip(clockwise_names, clockwise_features,
                                     clockwise_origins):
      player.walker.observables.add_egocentric_vector(
          name, base_observable.Generic(feature), origin_callable=origin)

  def _add_player_stats_observables(self, task, player):
    """Add observables corresponding to game statistics.

    Args:
      task: A `soccer.Task` instance.
      player: A `Walker` instance to which observables will be added.
    """

    def _stats_vel_to_ball(physics):
      dir_ = (
          physics.bind(task.ball.geom).xpos -
          physics.bind(player.walker.root_body).xpos)
      vel_to_ball = np.dot(dir_[:2] / (np.linalg.norm(dir_[:2]) + 1e-7),
                           physics.bind(player.walker.root_body).cvel[3:5])
      return np.sum(vel_to_ball)

    player.walker.observables.add_observable(
        'stats_vel_to_ball', base_observable.Generic(_stats_vel_to_ball))

    def _stats_closest_vel_to_ball(physics):
      """Velocity to the ball if this walker is the team's closest."""
      closest = None
      min_team_dist_to_ball = np.inf
      for player_ in task.players:
        if player_.team == player.team:
          dist_to_ball = np.linalg.norm(
              physics.bind(task.ball.geom).xpos -
              physics.bind(player_.walker.root_body).xpos)
          if dist_to_ball < min_team_dist_to_ball:
            min_team_dist_to_ball = dist_to_ball
            closest = player_
      if closest is player:
        return _stats_vel_to_ball(physics)
      return 0.

    player.walker.observables.add_observable(
        'stats_closest_vel_to_ball',
        base_observable.Generic(_stats_closest_vel_to_ball))

    def _stats_veloc_forward(physics):
      """Player's forward velocity."""
      return player.walker.observables.veloc_forward(physics)

    player.walker.observables.add_observable(
        'stats_veloc_forward', base_observable.Generic(_stats_veloc_forward))

    def _stats_vel_ball_to_goal(physics):
      """Ball velocity towards opponents' goal."""
      if player.team == team_lib.Team.HOME:
        goal = task.arena.away_goal
      else:
        goal = task.arena.home_goal

      goal_center = (goal.upper + goal.lower) / 2.
      direction = goal_center - physics.bind(task.ball.geom).xpos
      ball_vel_observable = task.ball.observables.linear_velocity
      ball_vel = ball_vel_observable.observation_callable(physics)()

      norm_dir = np.linalg.norm(direction)
      normalized_dir = direction / norm_dir if norm_dir else direction
      return np.sum(np.dot(normalized_dir, ball_vel))

    player.walker.observables.add_observable(
        'stats_vel_ball_to_goal',
        base_observable.Generic(_stats_vel_ball_to_goal))

    def _stats_avg_teammate_dist(physics):
      """Compute average distance from `walker` to its teammates."""
      teammate_dists = []
      for other in task.players:
        if player is other:
          continue
        if other.team != player.team:
          continue
        dist = np.linalg.norm(
            physics.bind(player.walker.root_body).xpos -
            physics.bind(other.walker.root_body).xpos)
        teammate_dists.append(dist)
      return np.mean(teammate_dists) if teammate_dists else 0.

    player.walker.observables.add_observable(
        'stats_home_avg_teammate_dist',
        base_observable.Generic(_stats_avg_teammate_dist))

    def _stats_teammate_spread_out(physics):
      """Compute average distance from `walker` to its teammates."""
      return _stats_avg_teammate_dist(physics) > 5.

    player.walker.observables.add_observable(
        'stats_teammate_spread_out',
        base_observable.Generic(_stats_teammate_spread_out))

    def _stats_home_score(unused_physics):
      if (task.arena.detected_goal() and
          task.arena.detected_goal() == player.team):
        return 1.
      return 0.

    player.walker.observables.add_observable(
        'stats_home_score', base_observable.Generic(_stats_home_score))

    has_opponent = any([p.team != player.team for p in task.players])

    def _stats_away_score(unused_physics):
      if (has_opponent and task.arena.detected_goal() and
          task.arena.detected_goal() != player.team):
        return 1.
      return 0.

    player.walker.observables.add_observable(
        'stats_away_score', base_observable.Generic(_stats_away_score))


# TODO(b/124848293): add unit-test interception observables.
class InterceptionObservablesAdder(ObservablesAdder):
  """Adds obervables representing interception events.

  These observables represent events where this player received the ball from
  another player, or when an opponent intercepted the ball from this player's
  team. For each type of event there are three different thresholds applied to
  the distance travelled by the ball since it last made contact with a player
  (5, 10, or 15 meters).

  For example, on a given timestep `stats_i_received_ball_10m` will be 1 if
  * This player just made contact with the ball
  * The last player to have made contact with the ball was a different player
  * The ball travelled for at least 10 m since it last hit a player
  and 0 otherwise.

  Conversely, `stats_opponent_intercepted_ball_10m` will be 1 if:
  * An opponent just made contact with the ball
  * The last player to have made contact with the ball was on this player's team
  * The ball travelled for at least 10 m since it last hit a player
  """

  def __call__(self, task, player):
    """Adds observables to a player for the given task.

    Args:
      task: A `soccer.Task` instance.
      player: A `Walker` instance to which observables will be added.
    """

    def _stats_i_received_ball(unused_physics):
      if (task.ball.hit and task.ball.repossessed and
          task.ball.last_hit is player):
        return 1.
      return 0.

    player.walker.observables.add_observable(
        'stats_i_received_ball',
        base_observable.Generic(_stats_i_received_ball))

    def _stats_opponent_intercepted_ball(unused_physics):
      """Indicator on if an opponent intercepted the ball."""
      if (task.ball.hit and task.ball.intercepted and
          task.ball.last_hit.team != player.team):
        return 1.
      return 0.

    player.walker.observables.add_observable(
        'stats_opponent_intercepted_ball',
        base_observable.Generic(_stats_opponent_intercepted_ball))

    for dist in [5, 10, 15]:

      def _stats_i_received_ball_dist(physics, dist=dist):
        if (_stats_i_received_ball(physics) and
            task.ball.dist_between_last_hits is not None and
            task.ball.dist_between_last_hits > dist):
          return 1.
        return 0.

      player.walker.observables.add_observable(
          'stats_i_received_ball_%dm' % dist,
          base_observable.Generic(_stats_i_received_ball_dist))

      def _stats_opponent_intercepted_ball_dist(physics, dist=dist):
        if (_stats_opponent_intercepted_ball(physics) and
            task.ball.dist_between_last_hits is not None and
            task.ball.dist_between_last_hits > dist):
          return 1.
        return 0.

      player.walker.observables.add_observable(
          'stats_opponent_intercepted_ball_%dm' % dist,
          base_observable.Generic(_stats_opponent_intercepted_ball_dist))



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.composer.props.position_detector."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.
from absl.testing import absltest
from absl.testing import parameterized
from dm_control import composer
from dm_control.entities.props import position_detector
from dm_control.entities.props import primitive
import numpy as np


class PositionDetectorTest(parameterized.TestCase):

  def setUp(self):
    super(PositionDetectorTest, self).setUp()
    self.arena = composer.Arena()
    self.props = [
        primitive.Primitive(geom_type='sphere', size=(0.1,)),
        primitive.Primitive(geom_type='sphere', size=(0.1,))
    ]
    for prop in self.props:
      self.arena.add_free_entity(prop)
    self.task = composer.NullTask(self.arena)

  def assertDetected(self, entity, detector):
    if not self.inverted:
      self.assertIn(entity, detector.detected_entities)
    else:
      self.assertNotIn(entity, detector.detected_entities)

  def assertNotDetected(self, entity, detector):
    if not self.inverted:
      self.assertNotIn(entity, detector.detected_entities)
    else:
      self.assertIn(entity, detector.detected_entities)

  @parameterized.parameters(False, True)
  def test3DDetection(self, inverted):
    self.inverted = inverted

    detector_pos = np.array([0.3, 0.2, 0.1])
    detector_size = np.array([0.1, 0.2, 0.3])
    detector = position_detector.PositionDetector(
        pos=detector_pos, size=detector_size, inverted=inverted)
    detector.register_entities(*self.props)
    self.arena.attach(detector)
    env = composer.Environment(self.task)

    env.reset()
    self.assertNotDetected(self.props[0], detector)
    self.assertNotDetected(self.props[1], detector)

    def initialize_episode(physics, unused_random_state):
      for prop in self.props:
        prop.set_pose(physics, detector_pos)
    self.task.initialize_episode = initialize_episode
    env.reset()
    self.assertDetected(self.props[0], detector)
    self.assertDetected(self.props[1], detector)

    self.props[0].set_pose(env.physics, detector_pos - detector_size)
    env.step([])
    self.assertNotDetected(self.props[0], detector)
    self.assertDetected(self.props[1], detector)

    self.props[0].set_pose(env.physics, detector_pos - detector_size / 2)
    self.props[1].set_pose(env.physics, detector_pos + detector_size * 1.01)
    env.step([])
    self.assertDetected(self.props[0], detector)
    self.assertNotDetected(self.props[1], detector)

  @parameterized.parameters(False, True)
  def test2DDetection(self, inverted):
    self.inverted = inverted

    detector_pos = np.array([0.3, 0.2])
    detector_size = np.array([0.1, 0.2])
    detector = position_detector.PositionDetector(
        pos=detector_pos, size=detector_size, inverted=inverted)
    detector.register_entities(*self.props)
    self.arena.attach(detector)
    env = composer.Environment(self.task)

    env.reset()
    self.assertNotDetected(self.props[0], detector)
    self.assertNotDetected(self.props[1], detector)

    def initialize_episode(physics, unused_random_state):
      # In 2D mode, detection should occur no matter how large |z| is.
      self.props[0].set_pose(physics, [detector_pos[0], detector_pos[1], 1e+6])
      self.props[1].set_pose(physics, [detector_pos[0], detector_pos[1], -1e+6])
    self.task.initialize_episode = initialize_episode
    env.reset()
    self.assertDetected(self.props[0], detector)
    self.assertDetected(self.props[1], detector)

    self.props[0].set_pose(
        env.physics, [detector_pos[0] - detector_size[0], detector_pos[1], 0])
    env.step([])
    self.assertNotDetected(self.props[0], detector)
    self.assertDetected(self.props[1], detector)

    self.props[0].set_pose(
        env.physics, [detector_pos[0] - detector_size[0] / 2,
                      detector_pos[1] + detector_size[1] / 2, 0])
    self.props[1].set_pose(
        env.physics, [detector_pos[0], detector_pos[1] + detector_size[1], 0])
    env.step([])
    self.assertDetected(self.props[0], detector)
    self.assertNotDetected(self.props[1], detector)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for the CMU humanoid."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
from dm_control import mjcf
from dm_control.composer.observation.observable import base as observable_base
from dm_control.locomotion.walkers import cmu_humanoid
import numpy as np
from six.moves import range
from six.moves import zip


class CMUHumanoidTest(parameterized.TestCase):

  @parameterized.parameters([
      cmu_humanoid.CMUHumanoid,
      cmu_humanoid.CMUHumanoidPositionControlled,
  ])
  def test_can_compile_and_step_simulation(self, walker_type):
    walker = walker_type()
    physics = mjcf.Physics.from_mjcf_model(walker.mjcf_model)
    for _ in range(100):
      physics.step()

  @parameterized.parameters([
      cmu_humanoid.CMUHumanoid,
      cmu_humanoid.CMUHumanoidPositionControlled,
  ])
  def test_actuators_sorted_alphabetically(self, walker_type):
    walker = walker_type()
    actuator_names = [
        actuator.name for actuator in walker.mjcf_model.find_all('actuator')]
    np.testing.assert_array_equal(actuator_names, sorted(actuator_names))

  def test_actuator_to_mocap_joint_mapping(self):
    walker = cmu_humanoid.CMUHumanoid()

    with self.subTest('Forward mapping'):
      for actuator_num, cmu_mocap_joint_num in enumerate(walker.actuator_order):
        self.assertEqual(walker.actuator_to_joint_order[cmu_mocap_joint_num],
                         actuator_num)

    with self.subTest('Inverse mapping'):
      for cmu_mocap_joint_num, actuator_num in enumerate(
          walker.actuator_to_joint_order):
        self.assertEqual(walker.actuator_order[actuator_num],
                         cmu_mocap_joint_num)

  def test_cmu_humanoid_position_controlled_has_correct_actuators(self):
    walker_torque = cmu_humanoid.CMUHumanoid()
    walker_pos = cmu_humanoid.CMUHumanoidPositionControlled()

    actuators_torque = walker_torque.mjcf_model.find_all('actuator')
    actuators_pos = walker_pos.mjcf_model.find_all('actuator')

    actuator_pos_params = {
        params.name: params for params in cmu_humanoid._POSITION_ACTUATORS}

    self.assertEqual(len(actuators_torque), len(actuators_pos))

    for actuator_torque, actuator_pos in zip(actuators_torque, actuators_pos):
      self.assertEqual(actuator_pos.name, actuator_torque.name)
      self.assertEqual(actuator_pos.joint.full_identifier,
                       actuator_torque.joint.full_identifier)
      self.assertEqual(actuator_pos.tag, 'general')
      self.assertEqual(actuator_pos.ctrllimited, 'true')
      np.testing.assert_array_equal(actuator_pos.ctrlrange, (-1, 1))

      expected_params = actuator_pos_params[actuator_pos.name]
      self.assertEqual(actuator_pos.biasprm[1], -expected_params.kp)
      np.testing.assert_array_equal(actuator_pos.forcerange,
                                    expected_params.forcerange)

  @parameterized.parameters([
      'body_camera',
      'egocentric_camera',
      'head',
      'left_arm_root',
      'right_arm_root',
      'root_body',
  ])
  def test_get_element_property(self, name):
    attribute_value = getattr(cmu_humanoid.CMUHumanoid(), name)
    self.assertIsInstance(attribute_value, mjcf.Element)

  @parameterized.parameters([
      'actuators',
      'bodies',
      'end_effectors',
      'marker_geoms',
      'mocap_joints',
      'observable_joints',
  ])
  def test_get_element_tuple_property(self, name):
    attribute_value = getattr(cmu_humanoid.CMUHumanoid(), name)
    self.assertNotEmpty(attribute_value)
    for item in attribute_value:
      self.assertIsInstance(item, mjcf.Element)

  def test_set_name(self):
    name = 'fred'
    walker = cmu_humanoid.CMUHumanoid(name=name)
    self.assertEqual(walker.mjcf_model.model, name)

  def test_set_marker_rgba(self):
    marker_rgba = (1., 0., 1., 0.5)
    walker = cmu_humanoid.CMUHumanoid(marker_rgba=marker_rgba)
    for marker_geom in walker.marker_geoms:
      np.testing.assert_array_equal(marker_geom.rgba, marker_rgba)

  @parameterized.parameters(
      'actuator_activation',
      'appendages_pos',
      'body_camera',
      'head_height',
      'sensors_torque',
  )
  def test_evaluate_observable(self, name):
    walker = cmu_humanoid.CMUHumanoid()
    observable = getattr(walker.observables, name)
    physics = mjcf.Physics.from_mjcf_model(walker.mjcf_model)
    observation = observable(physics)
    self.assertIsInstance(observation, (float, np.ndarray))

  def test_proprioception(self):
    walker = cmu_humanoid.CMUHumanoid()
    for item in walker.observables.proprioception:
      self.assertIsInstance(item, observable_base.Observable)

  def test_cmu_pose_to_actuation(self):
    walker = cmu_humanoid.CMUHumanoidPositionControlled()
    random_state = np.random.RandomState(123)

    expected_actuation = random_state.uniform(-1, 1, len(walker.actuator_order))

    cmu_limits = zip(*(joint.range for joint in walker.mocap_joints))
    cmu_lower, cmu_upper = (np.array(limit) for limit in cmu_limits)
    cmu_pose = cmu_lower + (cmu_upper - cmu_lower) * (
        1 + expected_actuation[walker.actuator_to_joint_order]) / 2

    actual_actuation = walker.cmu_pose_to_actuation(cmu_pose)

    np.testing.assert_allclose(actual_actuation, expected_actuation)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Escape locomotion tasks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control import composer
from dm_control import mjcf
from dm_control.composer.observation import observable as base_observable
from dm_control.rl import control
from dm_control.utils import rewards

import numpy as np

# Constants related to terrain generation.
_HEIGHTFIELD_ID = 0


class Escape(composer.Task):
  """A task solved by escaping a starting area (e.g. bowl-shaped terrain)."""

  def __init__(self,
               walker,
               arena,
               walker_spawn_position=(0, 0, 0),
               walker_spawn_rotation=None,
               physics_timestep=0.005,
               control_timestep=0.025):
    """Initializes this task.

    Args:
      walker: an instance of `locomotion.walkers.base.Walker`.
      arena: an instance of `locomotion.arenas`.
      walker_spawn_position: a sequence of 3 numbers, or a `composer.Variation`
        instance that generates such sequences, specifying the position at
        which the walker is spawned at the beginning of an episode.
      walker_spawn_rotation: a number, or a `composer.Variation` instance that
        generates a number, specifying the yaw angle offset (in radians) that is
        applied to the walker at the beginning of an episode.
      physics_timestep: a number specifying the timestep (in seconds) of the
        physics simulation.
      control_timestep: a number specifying the timestep (in seconds) at which
        the agent applies its control inputs (in seconds).
    """

    self._arena = arena
    self._walker = walker
    self._walker.create_root_joints(self._arena.attach(self._walker))
    self._walker_spawn_position = walker_spawn_position
    self._walker_spawn_rotation = walker_spawn_rotation

    enabled_observables = []
    enabled_observables += self._walker.observables.proprioception
    enabled_observables += self._walker.observables.kinematic_sensors
    enabled_observables += self._walker.observables.dynamic_sensors
    enabled_observables.append(self._walker.observables.sensors_touch)
    enabled_observables.append(self._walker.observables.egocentric_camera)
    for observable in enabled_observables:
      observable.enabled = True

    if 'CMUHumanoid' in str(type(self._walker)):
      core_body = 'walker/root'
      self._reward_body = 'walker/root'
    elif 'Rat' in str(type(self._walker)):
      core_body = 'walker/torso'
      self._reward_body = 'walker/head'
    else:
      raise ValueError('Expects Rat or CMUHumanoid.')

    def _origin(physics):
      """Returns origin position in the torso frame."""
      torso_frame = physics.named.data.xmat[core_body].reshape(3, 3)
      torso_pos = physics.named.data.xpos[core_body]
      return -torso_pos.dot(torso_frame)

    self._walker.observables.add_observable(
        'origin', base_observable.Generic(_origin))

    self.set_timesteps(
        physics_timestep=physics_timestep, control_timestep=control_timestep)

  @property
  def root_entity(self):
    return self._arena

  def initialize_episode_mjcf(self, random_state):
    if hasattr(self._arena, 'regenerate'):
      self._arena.regenerate(random_state)
    self._arena.mjcf_model.visual.map.znear = 0.00025
    self._arena.mjcf_model.visual.map.zfar = 50.

  def initialize_episode(self, physics, random_state):
    super(Escape, self).initialize_episode(physics, random_state)

    # Initial configuration.
    orientation = random_state.randn(4)
    orientation /= np.linalg.norm(orientation)
    _find_non_contacting_height(physics, self._walker, orientation)

  def get_reward(self, physics):
    # Escape reward term.
    terrain_size = physics.model.hfield_size[_HEIGHTFIELD_ID, 0]
    escape_reward = rewards.tolerance(
        np.asarray(np.linalg.norm(
            physics.named.data.site_xpos[self._reward_body])),
        bounds=(terrain_size, float('inf')),
        margin=terrain_size,
        value_at_margin=0,
        sigmoid='linear')
    upright_reward = _upright_reward(physics, self._walker, deviation_angle=30)
    return upright_reward * escape_reward

  def get_discount(self, physics):
    return 1.


def _find_non_contacting_height(physics, walker, orientation,
                                x_pos=0.0, y_pos=0.0, maxiter=1000):
  """Find a height with no contacts given a body orientation.

  Args:
    physics: An instance of `Physics`.
    walker: the focal walker.
    orientation: A quaternion.
    x_pos: A float. Position along global x-axis.
    y_pos: A float. Position along global y-axis.
    maxiter: maximum number of iterations to try
  """
  z_pos = 0.0  # Start embedded in the floor.
  num_contacts = 1
  count = 1
  # Move up in 1cm increments until no contacts.
  while num_contacts > 0:
    try:
      with physics.reset_context():
        freejoint = mjcf.get_frame_freejoint(walker.mjcf_model)
        physics.bind(freejoint).qpos[:3] = x_pos, y_pos, z_pos
        physics.bind(freejoint).qpos[3:] = orientation
    except control.PhysicsError:
      # We may encounter a PhysicsError here due to filling the contact
      # buffer, in which case we simply increment the height and continue.
      pass
    num_contacts = physics.data.ncon
    z_pos += 0.01
    count += 1
    if count > maxiter:
      raise ValueError(
          'maxiter reached: possibly contacts in null pose of body.'
      )


def _upright_reward(physics, walker, deviation_angle=0):
  """Returns a reward proportional to how upright the torso is.

  Args:
    physics: an instance of `Physics`.
    walker: the focal walker.
    deviation_angle: A float, in degrees. The reward is 0 when the torso is
      exactly upside-down and 1 when the torso's z-axis is less than
      `deviation_angle` away from the global z-axis.
  """
  deviation = np.cos(np.deg2rad(deviation_angle))
  upright_torso = physics.bind(walker.root_body).xmat[-1]
  if hasattr(walker, 'pelvis_body'):
    upright_pelvis = physics.bind(walker.pelvis_body).xmat[-1]
    upright_zz = np.stack([upright_torso, upright_pelvis])
  else:
    upright_zz = upright_torso
  upright = rewards.tolerance(upright_zz,
                              bounds=(deviation, float('inf')),
                              sigmoid='linear',
                              margin=1 + deviation,
                              value_at_margin=0)
  return np.min(upright)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Main user-facing classes and utility functions for loading MuJoCo models."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
import ctypes
import os
import weakref

from absl import logging

from dm_control.mujoco.wrapper import util
from dm_control.mujoco.wrapper.mjbindings import constants
from dm_control.mujoco.wrapper.mjbindings import enums
from dm_control.mujoco.wrapper.mjbindings import functions
from dm_control.mujoco.wrapper.mjbindings import mjlib
from dm_control.mujoco.wrapper.mjbindings import types
from dm_control.mujoco.wrapper.mjbindings import wrappers
import numpy as np
import six

# Internal analytics import.
# Unused internal import: resources.

_NULL = b"\00"
_FAKE_XML_FILENAME = b"model.xml"
_FAKE_BINARY_FILENAME = b"model.mjb"

# Although `mjMAXVFSNAME` from `mjmodel.h` specifies a limit of 100 bytes
# (including the terminal null byte), the actual limit seems to be 99 bytes
# (98 characters).
_MAX_VFS_FILENAME_CHARACTERS = 98
_VFS_FILENAME_TOO_LONG = (
    "Filename length {length} exceeds {limit} character limit: {filename}")
_INVALID_FONT_SCALE = ("`font_scale` must be one of {}, got {{}}."
                       .format(enums.mjtFontScale))

_CONTACT_ID_OUT_OF_RANGE = (
    "`contact_id` must be between 0 and {max_valid} (inclusive), got: {actual}."
)

# Global cache used to store finalizers for freeing ctypes pointers.
# Contains {pointer_address: weakref_object} pairs.
_FINALIZERS = {}


class Error(Exception):
  """Base class for MuJoCo exceptions."""
  pass


if constants.mjVERSION_HEADER != mjlib.mj_version():
  raise Error("MuJoCo library version ({0}) does not match header version "
              "({1})".format(constants.mjVERSION_HEADER, mjlib.mj_version()))

_REGISTERED = False
_ERROR_BUFSIZE = 1000

# This is used to keep track of the `MJMODEL` pointer that was most recently
# loaded by `_get_model_ptr_from_xml`. Only this model can be saved to XML.
_LAST_PARSED_MODEL_PTR = None

_NOT_LAST_PARSED_ERROR = (
    "Only the model that was most recently loaded from an XML file or string "
    "can be saved to an XML file.")

import time

# NB: Python functions that are called from C are defined at module-level to
# ensure they won't be garbage-collected before they are called.
@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _warning_callback(message):
  logging.warning(util.to_native_string(message))


@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _error_callback(message):
  logging.fatal(util.to_native_string(message))


# Override MuJoCo's callbacks for handling warnings and errors.
mjlib.mju_user_warning = ctypes.c_void_p.in_dll(mjlib, "mju_user_warning")
mjlib.mju_user_error = ctypes.c_void_p.in_dll(mjlib, "mju_user_error")
mjlib.mju_user_warning.value = ctypes.cast(
    _warning_callback, ctypes.c_void_p).value
mjlib.mju_user_error.value = ctypes.cast(
    _error_callback, ctypes.c_void_p).value


def enable_timer(enabled=True):
  if enabled:
    set_callback("mjcb_time", time.time)
  else:
    set_callback("mjcb_time", None)


def _maybe_register_license(path=None):
  """Registers the MuJoCo license if not already registered.

  Args:
    path: Optional custom path to license key file.

  Raises:
    Error: If the license could not be registered.
  """
  global _REGISTERED
  if not _REGISTERED:
    if path is None:
      path = util.get_mjkey_path()
    result = mjlib.mj_activate(util.to_binary_string(path))
    if result == 1:
      _REGISTERED = True
      # Internal analytics of mj_activate.
    elif result == 0:
      raise Error("Could not register license.")
    else:
      raise Error("Unknown registration error (code: {})".format(result))


def _str2type(type_str):
  type_id = mjlib.mju_str2Type(util.to_binary_string(type_str))
  if not type_id:
    raise Error("{!r} is not a valid object type name.".format(type_str))
  return type_id


def _type2str(type_id):
  type_str_ptr = mjlib.mju_type2Str(type_id)
  if not type_str_ptr:
    raise Error("{!r} is not a valid object type ID.".format(type_id))
  return ctypes.string_at(type_str_ptr)


def set_callback(name, new_callback=None):
  """Sets a user-defined callback function to modify MuJoCo's behavior.

  Callback functions should have the following signature:
    func(const_mjmodel_ptr, mjdata_ptr) -> None

  Args:
    name: Name of the callback to set. Must be a field in
      `functions.function_pointers`.
    new_callback: The new callback. This can be one of the following:
      * A Python callable
      * A C function exposed by a `ctypes.CDLL` object
      * An integer specifying the address of a callback function
      * None, in which case any existing callback of that name is removed
  """
  setattr(functions.callbacks, name, new_callback)


@contextlib.contextmanager
def callback_context(name, new_callback=None):
  """Context manager that temporarily overrides a MuJoCo callback function.

  On exit, the callback will be restored to its original value (None if the
  callback was not already overridden when the context was entered).

  Args:
    name: Name of the callback to set. Must be a field in
      `mjbindings.function_pointers`.
    new_callback: The new callback. This can be one of the following:
      * A Python callable
      * A C function exposed by a `ctypes.CDLL` object
      * An integer specifying the address of a callback function
      * None, in which case any existing callback of that name is removed

  Yields:
    None
  """
  old_callback = getattr(functions.callbacks, name)
  set_callback(name, new_callback)
  try:
    yield
  finally:
    # Ensure that the callback is reset on exit, even if an exception is raised.
    set_callback(name, old_callback)


def get_schema():
  """Returns a string containing the schema used by the MuJoCo XML parser."""
  buf = ctypes.create_string_buffer(100000)
  mjlib.mj_printSchema(None, buf, len(buf), 0, 0)
  return buf.value


@contextlib.contextmanager
def _temporary_vfs(filenames_and_contents):
  """Creates a temporary VFS containing one or more files.

  Args:
    filenames_and_contents: A dict containing `{filename: contents}` pairs.
      The length of each filename must not exceed 98 characters.

  Yields:
    A `types.MJVFS` instance.

  Raises:
    Error: If a file cannot be added to the VFS, or if an error occurs when
      looking up the filename.
    ValueError: If the length of a filename exceeds 98 characters.
  """
  vfs = types.MJVFS()
  mjlib.mj_defaultVFS(vfs)
  for filename, contents in six.iteritems(filenames_and_contents):
    if len(filename) > _MAX_VFS_FILENAME_CHARACTERS:
      raise ValueError(
          _VFS_FILENAME_TOO_LONG.format(
              length=len(filename),
              limit=_MAX_VFS_FILENAME_CHARACTERS,
              filename=filename))
    filename = util.to_binary_string(filename)
    contents = util.to_binary_string(contents)
    _, extension = os.path.splitext(filename)
    # For XML files we need to append a NULL byte, otherwise MuJoCo's parser
    # can sometimes read past the end of the string. However, we should *not*
    # do this for other file types (in particular for STL meshes, where this
    # causes MuJoCo's compiler to complain that the file size is incorrect).
    append_null = extension.lower() == b".xml"
    num_bytes = len(contents) + append_null
    retcode = mjlib.mj_makeEmptyFileVFS(vfs, filename, num_bytes)
    if retcode == 1:
      raise Error("Failed to create {!r}: VFS is full.".format(filename))
    elif retcode == 2:
      raise Error("Failed to create {!r}: duplicate filename.".format(filename))
    file_index = mjlib.mj_findFileVFS(vfs, filename)
    if file_index == -1:
      raise Error("Could not find {!r} in the VFS".format(filename))
    vf = vfs.filedata[file_index]
    vf_as_char_arr = ctypes.cast(vf, ctypes.POINTER(ctypes.c_char * num_bytes))
    vf_as_char_arr.contents[:len(contents)] = contents
    if append_null:
      vf_as_char_arr.contents[-1] = _NULL
  try:
    yield vfs
  finally:
    mjlib.mj_deleteVFS(vfs)  # Ensure that we free the VFS afterwards.


def _create_finalizer(ptr, free_func):
  """Creates a finalizer for a ctypes pointer.

  Args:
    ptr: A `ctypes.POINTER` to be freed.
    free_func: A callable that frees the pointer. It will be called with `ptr`
      as its only argument when `ptr` is garbage collected.
  """
  ptr_type = type(ptr)
  address = ctypes.addressof(ptr.contents)

  if address not in _FINALIZERS:  # Only one finalizer needed per address.

    logging.debug("Allocated %s at %x", ptr_type.__name__, address)

    def callback(dead_ptr_ref):
      """A weakref callback that frees the resource held by a pointer."""
      del dead_ptr_ref  # Unused weakref to the dead ctypes pointer object.
      if address not in _FINALIZERS:
        # Someone had already explicitly called `call_finalizer_for_pointer`.
        return
      else:
        # Turn the address back into a pointer to be freed.
        temp_ptr = ctypes.cast(address, ptr_type)
        free_func(temp_ptr)
        logging.debug("Freed %s at %x", ptr_type.__name__, address)
        del _FINALIZERS[address]  # Remove the weakref from the global cache.

    # Store weakrefs in a global cache so that they don't get garbage collected
    # before their referents.
    _FINALIZERS[address] = (weakref.ref(ptr, callback), callback)


def _finalize(ptr):
  """Calls the finalizer for the specified pointer to free allocated memory."""
  address = ctypes.addressof(ptr.contents)
  try:
    ptr_ref, callback = _FINALIZERS[address]
    callback(ptr_ref)
  except KeyError:
    pass


def _load_xml(filename, vfs_or_none):
  """Invokes `mj_loadXML` with logging/error handling."""
  error_buf = ctypes.create_string_buffer(_ERROR_BUFSIZE)
  model_ptr = mjlib.mj_loadXML(
      util.to_binary_string(filename),
      vfs_or_none,
      error_buf,
      _ERROR_BUFSIZE)
  if not model_ptr:
    raise Error(util.to_native_string(error_buf.value))
  elif error_buf.value:
    logging.warning(util.to_native_string(error_buf.value))

  # Free resources when the ctypes pointer is garbage collected.
  _create_finalizer(model_ptr, mjlib.mj_deleteModel)

  return model_ptr


def _get_model_ptr_from_xml(xml_path=None, xml_string=None, assets=None):
  """Parses a model XML file, compiles it, and returns a pointer to an mjModel.

  Args:
    xml_path: None or a path to a model XML file in MJCF or URDF format.
    xml_string: None or an XML string containing an MJCF or URDF model
      description.
    assets: None or a dict containing external assets referenced by the model
      (such as additional XML files, textures, meshes etc.), in the form of
      `{filename: contents_string}` pairs. The keys should correspond to the
      filenames specified in the model XML. Ignored if `xml_string` is None.

    One of `xml_path` or `xml_string` must be specified.

  Returns:
    A `ctypes.POINTER` to a new `mjbindings.types.MJMODEL` instance.

  Raises:
    TypeError: If both or neither of `xml_path` and `xml_string` are specified.
    Error: If the model is not created successfully.
  """
  if xml_path is None and xml_string is None:
    raise TypeError(
        "At least one of `xml_path` or `xml_string` must be specified.")
  elif xml_path is not None and xml_string is not None:
    raise TypeError(
        "Only one of `xml_path` or `xml_string` may be specified.")

  _maybe_register_license()

  if xml_string is not None:
    assets = {} if assets is None else assets.copy()
    # Ensure that the fake XML filename doesn't overwrite an existing asset.
    xml_path = _FAKE_XML_FILENAME
    while xml_path in assets:
      xml_path = "_" + xml_path
    assets[xml_path] = xml_string
    with _temporary_vfs(assets) as vfs:
      ptr = _load_xml(xml_path, vfs)
  else:
    ptr = _load_xml(xml_path, None)

  global _LAST_PARSED_MODEL_PTR
  _LAST_PARSED_MODEL_PTR = ptr

  return ptr


def save_last_parsed_model_to_xml(xml_path, check_model=None):
  """Writes a description of the most recently loaded model to an MJCF XML file.

  Args:
    xml_path: Path to the output XML file.
    check_model: Optional `MjModel` instance. If specified, this model will be
      checked to see if it is the most recently parsed one, and a ValueError
      will be raised otherwise.
  Raises:
    Error: If MuJoCo encounters an error while writing the XML file.
    ValueError: If `check_model` was passed, and this model is not the most
      recently parsed one.
  """
  if check_model and check_model.ptr is not _LAST_PARSED_MODEL_PTR:
    raise ValueError(_NOT_LAST_PARSED_ERROR)
  error_buf = ctypes.create_string_buffer(_ERROR_BUFSIZE)
  mjlib.mj_saveLastXML(util.to_binary_string(xml_path),
                       _LAST_PARSED_MODEL_PTR,
                       error_buf,
                       _ERROR_BUFSIZE)
  if error_buf.value:
    raise Error(error_buf.value)


def _get_model_ptr_from_binary(binary_path=None, byte_string=None):
  """Returns a pointer to an mjModel from the contents of a MuJoCo model binary.

  Args:
    binary_path: Path to an MJB file (as produced by MjModel.save_binary).
    byte_string: String of bytes (as returned by MjModel.to_bytes).

    One of `binary_path` or `byte_string` must be specified.

  Returns:
    A `ctypes.POINTER` to a new `mjbindings.types.MJMODEL` instance.

  Raises:
    TypeError: If both or neither of `byte_string` and `binary_path`
      are specified.
  """
  if binary_path is None and byte_string is None:
    raise TypeError(
        "At least one of `byte_string` or `binary_path` must be specified.")
  elif binary_path is not None and byte_string is not None:
    raise TypeError(
        "Only one of `byte_string` or `binary_path` may be specified.")

  _maybe_register_license()

  if byte_string is not None:
    with _temporary_vfs({_FAKE_BINARY_FILENAME: byte_string}) as vfs:
      ptr = mjlib.mj_loadModel(_FAKE_BINARY_FILENAME, vfs)
  else:
    ptr = mjlib.mj_loadModel(util.to_binary_string(binary_path), None)

  # Free resources when the ctypes pointer is garbage collected.
  _create_finalizer(ptr, mjlib.mj_deleteModel)

  return ptr


# Subclasses implementing constructors/destructors for low-level wrappers.
# ------------------------------------------------------------------------------


class MjModel(wrappers.MjModelWrapper):
  """Wrapper class for a MuJoCo 'mjModel' instance.

  MjModel encapsulates features of the model that are expected to remain
  constant. It also contains simulation and visualization options which may be
  changed occasionally, although this is done explicitly by the user.
  """

  def __init__(self, model_ptr):
    """Creates a new MjModel instance from a ctypes pointer.

    Args:
      model_ptr: A `ctypes.POINTER` to an `mjbindings.types.MJMODEL` instance.
    """
    super(MjModel, self).__init__(ptr=model_ptr)

  def __getstate__(self):
    # All of MjModel's state is assumed to reside within the MuJoCo C struct.
    # However there is no mechanism to prevent users from adding arbitrary
    # Python attributes to an MjModel instance - these would not be serialized.
    return self.to_bytes()

  def __setstate__(self, byte_string):
    model_ptr = _get_model_ptr_from_binary(byte_string=byte_string)
    self.__init__(model_ptr)

  def __copy__(self):
    new_model_ptr = mjlib.mj_copyModel(None, self.ptr)
    return self.__class__(new_model_ptr)

  @classmethod
  def from_xml_string(cls, xml_string, assets=None):
    """Creates an `MjModel` instance from a model description XML string.

    Args:
      xml_string: String containing an MJCF or URDF model description.
      assets: Optional dict containing external assets referenced by the model
        (such as additional XML files, textures, meshes etc.), in the form of
        `{filename: contents_string}` pairs. The keys should correspond to the
        filenames specified in the model XML.

    Returns:
      An `MjModel` instance.
    """
    model_ptr = _get_model_ptr_from_xml(xml_string=xml_string, assets=assets)
    return cls(model_ptr)

  @classmethod
  def from_byte_string(cls, byte_string):
    """Creates an MjModel instance from a model binary as a string of bytes."""
    model_ptr = _get_model_ptr_from_binary(byte_string=byte_string)
    return cls(model_ptr)

  @classmethod
  def from_xml_path(cls, xml_path):
    """Creates an MjModel instance from a path to a model XML file."""
    model_ptr = _get_model_ptr_from_xml(xml_path=xml_path)
    return cls(model_ptr)

  @classmethod
  def from_binary_path(cls, binary_path):
    """Creates an MjModel instance from a path to a compiled model binary."""
    model_ptr = _get_model_ptr_from_binary(binary_path=binary_path)
    return cls(model_ptr)

  def save_binary(self, binary_path):
    """Saves the MjModel instance to a binary file."""
    mjlib.mj_saveModel(self.ptr, util.to_binary_string(binary_path), None, 0)

  def to_bytes(self):
    """Serialize the model to a string of bytes."""
    bufsize = mjlib.mj_sizeModel(self.ptr)
    buf = ctypes.create_string_buffer(bufsize)
    mjlib.mj_saveModel(self.ptr, None, buf, bufsize)
    return buf.raw

  def copy(self):
    """Returns a copy of this MjModel instance."""
    return self.__copy__()

  def free(self):
    """Frees the native resources held by this MjModel.

    This is an advanced feature for use when manual memory management is
    necessary. This MjModel object MUST NOT be used after this function has
    been called.
    """
    _finalize(self._ptr)
    del self._ptr

  def name2id(self, name, object_type):
    """Returns the integer ID of a specified MuJoCo object.

    Args:
      name: String specifying the name of the object to query.
      object_type: The type of the object. Can be either a lowercase string
        (e.g. 'body', 'geom') or an `mjtObj` enum value.

    Returns:
      An integer object ID.

    Raises:
      Error: If `object_type` is not a valid MuJoCo object type, or if no object
        with the corresponding name and type was found.
    """
    if not isinstance(object_type, int):
      object_type = _str2type(object_type)
    obj_id = mjlib.mj_name2id(
        self.ptr, object_type, util.to_binary_string(name))
    if obj_id == -1:
      raise Error("Object of type {!r} with name {!r} does not exist.".format(
          _type2str(object_type), name))
    return obj_id

  def id2name(self, object_id, object_type):
    """Returns the name associated with a MuJoCo object ID, if there is one.

    Args:
      object_id: Integer ID.
      object_type: The type of the object. Can be either a lowercase string
        (e.g. 'body', 'geom') or an `mjtObj` enum value.

    Returns:
      A string containing the object name, or an empty string if the object ID
      either doesn't exist or has no name.

    Raises:
      Error: If `object_type` is not a valid MuJoCo object type.
    """
    if not isinstance(object_type, int):
      object_type = _str2type(object_type)
    name_ptr = mjlib.mj_id2name(self.ptr, object_type, object_id)
    if not name_ptr:
      return ""
    return util.to_native_string(ctypes.string_at(name_ptr))

  @contextlib.contextmanager
  def disable(self, *flags):
    """Context manager for temporarily disabling MuJoCo flags.

    Args:
      *flags: Positional arguments specifying flags to disable. Can be either
        lowercase strings (e.g. 'gravity', 'contact') or `mjtDisableBit` enum
        values.

    Yields:
      None

    Raises:
      ValueError: If any item in `flags` is neither a valid name nor a value
        from `enums.mjtDisableBit`.
    """
    old_bitmask = self.opt.disableflags
    new_bitmask = old_bitmask
    for flag in flags:
      if isinstance(flag, six.string_types):
        try:
          field_name = "mjDSBL_" + flag.upper()
          bitmask = getattr(enums.mjtDisableBit, field_name)
        except AttributeError:
          valid_names = [field_name.split("_")[1].lower()
                         for field_name in enums.mjtDisableBit._fields[:-1]]
          raise ValueError("'{}' is not a valid flag name. Valid names: {}"
                           .format(flag, ", ".join(valid_names)))
      else:
        if flag not in enums.mjtDisableBit[:-1]:
          raise ValueError("'{}' is not a value in `enums.mjtDisableBit`. "
                           "Valid values: {}"
                           .format(flag, tuple(enums.mjtDisableBit[:-1])))
        bitmask = flag
      new_bitmask |= bitmask
    self.opt.disableflags = new_bitmask
    try:
      yield
    finally:
      self.opt.disableflags = old_bitmask

  @property
  def name(self):
    """Returns the name of the model."""
    # The model name is the first null-terminated string in the `names` buffer.
    return util.to_native_string(
        ctypes.string_at(ctypes.addressof(self.names.contents)))


class MjData(wrappers.MjDataWrapper):
  """Wrapper class for a MuJoCo 'mjData' instance.

  MjData contains all of the dynamic variables and intermediate results produced
  by the simulation. These are expected to change on each simulation timestep.
  """

  def __init__(self, model):
    """Construct a new MjData instance.

    Args:
      model: An MjModel instance.
    """
    self._model = model

    # Allocate resources for mjData.
    data_ptr = mjlib.mj_makeData(model.ptr)

    # Free resources when the ctypes pointer is garbage collected.
    _create_finalizer(data_ptr, mjlib.mj_deleteData)

    super(MjData, self).__init__(data_ptr, model)

  def __getstate__(self):
    # Note: we can replace this once a `saveData` MJAPI function exists.
    # To reconstruct an MjData instance we need three things:
    #   1. Its parent MjModel instance
    #   2. A subset of its fixed-size fields whose values aren't determined by
    #      the model
    #   3. The contents of its internal buffer (all of its pointer fields point
    #      into this)
    struct_fields = {}
    for name in ["solver", "timer", "warning"]:
      struct_fields[name] = getattr(self, name).copy()
    scalar_field_names = ["ncon", "time", "energy"]
    scalar_fields = {name: getattr(self, name) for name in scalar_field_names}
    static_fields = {"struct_fields": struct_fields,
                     "scalar_fields": scalar_fields}
    buffer_contents = ctypes.string_at(self.buffer_, self.nbuffer)
    return (self._model, static_fields, buffer_contents)

  def __setstate__(self, state_tuple):
    # Replace this once a `loadData` MJAPI function exists.
    self._model, static_fields, buffer_contents = state_tuple
    self.__init__(self.model)
    for name, contents in six.iteritems(static_fields["struct_fields"]):
      getattr(self, name)[:] = contents

    for name, value in six.iteritems(static_fields["scalar_fields"]):
      # Array and scalar values must be handled separately.
      try:
        getattr(self, name)[:] = value
      except TypeError:
        setattr(self, name, value)
    buf_ptr = (ctypes.c_char * self.nbuffer).from_address(self.buffer_)
    buf_ptr[:] = buffer_contents

  def __copy__(self):
    # This makes a shallow copy that shares the same parent MjModel instance.
    new_obj = self.__class__(self.model)
    mjlib.mj_copyData(new_obj.ptr, self.model.ptr, self.ptr)
    return new_obj

  def copy(self):
    """Returns a copy of this MjData instance with the same parent MjModel."""
    return self.__copy__()

  def free(self):
    """Frees the native resources held by this MjData.

    This is an advanced feature for use when manual memory management is
    necessary. This MjData object MUST NOT be used after this function has
    been called.
    """
    _finalize(self._ptr)
    del self._ptr

  def object_velocity(self, object_id, object_type, local_frame=False):
    """Returns the 6D velocity (linear, angular) of a MuJoCo object.

    Args:
      object_id: Object identifier. Can be either integer ID or String name.
      object_type: The type of the object. Can be either a lowercase string
        (e.g. 'body', 'geom') or an `mjtObj` enum value.
      local_frame: Boolean specifiying whether the velocity is given in the
        global (worldbody), or local (object) frame.

    Returns:
      2x3 array with stacked (linear_velocity, angular_velocity)

    Raises:
      Error: If `object_type` is not a valid MuJoCo object type, or if no object
        with the corresponding name and type was found.
    """
    if not isinstance(object_type, int):
      object_type = _str2type(object_type)
    velocity = np.empty(6, dtype=np.float64)
    if not isinstance(object_id, int):
      object_id = self.model.name2id(object_id, object_type)
    mjlib.mj_objectVelocity(self.model.ptr, self.ptr,
                            object_type, object_id, velocity, local_frame)
    #  MuJoCo returns velocities in (angular, linear) order, which we flip here.
    return velocity.reshape(2, 3)[::-1]

  def contact_force(self, contact_id):
    """Returns the wrench of a contact as a 2 x 3 array of (forces, torques).

    Args:
      contact_id: Integer, the index of the contact within the contact buffer
        (`self.contact`).

    Returns:
      2x3 array with stacked (force, torque). Note that the order of dimensions
        is (normal, tangent, tangent), in the contact's frame.

    Raises:
      ValueError: If `contact_id` is negative or bigger than ncon-1.
    """
    if not 0 <= contact_id < self.ncon:
      raise ValueError(_CONTACT_ID_OUT_OF_RANGE
                       .format(max_valid=self.ncon-1, actual=contact_id))
    wrench = np.empty(6, dtype=np.float64)
    mjlib.mj_contactForce(self.model.ptr, self.ptr, contact_id, wrench)
    return wrench.reshape(2, 3)

  @property
  def model(self):
    """The parent MjModel for this MjData instance."""
    return self._model

  @util.CachedProperty
  def _contact_buffer(self):
    """Cached structured array containing the full contact buffer."""
    contact_array = util.buf_to_npy(
        super(MjData, self).contact, shape=(self._model.nconmax,))
    return contact_array

  @property
  def contact(self):
    """Variable-length recarray containing all current contacts."""
    return self._contact_buffer[:self.ncon]


# Docstrings for these subclasses are inherited from their Wrapper parent class.


class MjvCamera(wrappers.MjvCameraWrapper):

  def __init__(self):
    ptr = ctypes.pointer(types.MJVCAMERA())
    mjlib.mjv_defaultCamera(ptr)
    super(MjvCamera, self).__init__(ptr)


class MjvOption(wrappers.MjvOptionWrapper):

  def __init__(self):
    ptr = ctypes.pointer(types.MJVOPTION())
    mjlib.mjv_defaultOption(ptr)
    # Do not visualize rangefinder lines by default:
    ptr.contents.flags[enums.mjtVisFlag.mjVIS_RANGEFINDER] = False
    super(MjvOption, self).__init__(ptr)


class UnmanagedMjrContext(wrappers.MjrContextWrapper):
  """A wrapper for MjrContext that does not manage the native object's lifetime.

  This wrapper is provided for API backward-compatibility reasons, since the
  creating and destruction of an MjrContext requires an OpenGL context to be
  provided.
  """

  def __init__(self):
    ptr = ctypes.pointer(types.MJRCONTEXT())
    mjlib.mjr_defaultContext(ptr)
    super(UnmanagedMjrContext, self).__init__(ptr)


class MjrContext(wrappers.MjrContextWrapper):  # pylint: disable=missing-docstring

  def __init__(self,
               model,
               gl_context,
               font_scale=enums.mjtFontScale.mjFONTSCALE_150):
    """Initializes this MjrContext instance.

    Args:
      model: An `MjModel` instance.
      gl_context: A `render.ContextBase` instance.
      font_scale: Integer controlling the font size for text. Must be a value
        in `mjbindings.enums.mjtFontScale`.

    Raises:
      ValueError: If `font_scale` is invalid.
    """
    if font_scale not in enums.mjtFontScale:
      raise ValueError(_INVALID_FONT_SCALE.format(font_scale))

    ptr = ctypes.pointer(types.MJRCONTEXT())
    mjlib.mjr_defaultContext(ptr)

    with gl_context.make_current() as ctx:
      ctx.call(mjlib.mjr_makeContext, model.ptr, ptr, font_scale)
      ctx.call(mjlib.mjr_setBuffer, enums.mjtFramebuffer.mjFB_OFFSCREEN, ptr)
      gl_context.increment_refcount()

    # Free resources when the ctypes pointer is garbage collected.
    def finalize_mjr_context(ptr):
      if not gl_context.terminated:
        with gl_context.make_current() as ctx:
          ctx.call(mjlib.mjr_freeContext, ptr)
          gl_context.decrement_refcount()

    _create_finalizer(ptr, finalize_mjr_context)

    super(MjrContext, self).__init__(ptr)

  def free(self):
    """Frees the native resources held by this MjrContext.

    This is an advanced feature for use when manual memory management is
    necessary. This MjrContext object MUST NOT be used after this function has
    been called.
    """
    _finalize(self._ptr)
    del self._ptr


# A mapping from human-readable short names to mjtRndFlag enum values, i.e.
# {'shadow': mjtRndFlag.mjRND_SHADOW, 'fog': mjtRndFlag.mjRND_FOG, ...}
_NAME_TO_RENDER_FLAG_ENUM_VALUE = {
    name[len("mjRND_"):].lower(): getattr(enums.mjtRndFlag, name)
    for name in enums.mjtRndFlag._fields[:-1]  # Exclude mjRND_NUMRNDFLAG entry.
}


class MjvScene(wrappers.MjvSceneWrapper):  # pylint: disable=missing-docstring

  def __init__(self, model=None, max_geom=1000):
    """Initializes a new `MjvScene` instance.

    Args:
      model: (optional) An `MjModel` instance.
      max_geom: (optional) An integer specifying the maximum number of geoms
        that can be represented in the scene.
    """
    model_ptr = model.ptr if model is not None else None
    scene_ptr = ctypes.pointer(types.MJVSCENE())

    # Allocate and initialize resources for the abstract scene.
    mjlib.mjv_makeScene(model_ptr, scene_ptr, max_geom)

    # Free resources when the ctypes pointer is garbage collected.
    _create_finalizer(scene_ptr, mjlib.mjv_freeScene)

    super(MjvScene, self).__init__(scene_ptr)

  @contextlib.contextmanager
  def override_flags(self, overrides):
    """Context manager for temporarily overriding rendering flags.

    Args:
      overrides: A mapping specifying rendering flags to override. The keys can
        be either lowercase strings or `mjtRndFlag` enum values, and the values
        are the overridden flag values, e.g. `{'wireframe': True}` or
        `{enums.mjtRndFlag.mjRND_WIREFRAME: True}`. See `enums.mjtRndFlag` for
        the set of valid flags.

    Yields:
      None
    """
    if not overrides:
      yield
    else:
      original_flags = self.flags.copy()
      for key, value in overrides.items():
        index = _NAME_TO_RENDER_FLAG_ENUM_VALUE.get(key, key)
        self.flags[index] = value
      try:
        yield
      finally:
        np.copyto(self.flags, original_flags)

  def free(self):
    """Frees the native resources held by this MjvScene.

    This is an advanced feature for use when manual memory management is
    necessary. This MjvScene object MUST NOT be used after this function has
    been called.
    """
    _finalize(self._ptr)
    del self._ptr

  @util.CachedProperty
  def _geoms_buffer(self):
    """Cached recarray containing the full geom buffer."""
    return util.buf_to_npy(super(MjvScene, self).geoms, shape=(self.maxgeom,))

  @property
  def geoms(self):
    """Variable-length recarray containing all geoms currently in the buffer."""
    return self._geoms_buffer[:self.ngeom]


class MjvPerturb(wrappers.MjvPerturbWrapper):

  def __init__(self):
    ptr = ctypes.pointer(types.MJVPERTURB())
    mjlib.mjv_defaultPerturb(ptr)
    super(MjvPerturb, self).__init__(ptr)


class MjvFigure(wrappers.MjvFigureWrapper):

  def __init__(self):
    ptr = ctypes.pointer(types.MJVFIGURE())
    mjlib.mjv_defaultFigure(ptr)
    super(MjvFigure, self).__init__(ptr)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for util."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import resource

# Internal dependencies.
from absl.testing import absltest
from dm_control.mujoco import wrapper
from dm_control.mujoco.wrapper import util
from six.moves import range

_NUM_CALLS = 100000
_RSS_GROWTH_TOLERANCE = 2000  # Bytes


class UtilTest(absltest.TestCase):

  def test_buf_to_npy_no_memory_leak(self):
    """Ensures we can call buf_to_npy without leaking memory."""
    model = wrapper.MjModel.from_xml_string("<mujoco/>")
    src = model._ptr.contents.name_geomadr
    shape = (model.ngeom,)

    # This uses high water marks to find memory leaks in native code.
    old_max = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    for _ in range(_NUM_CALLS):
      buf = util.buf_to_npy(src, shape)
    del buf
    new_max = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    growth = new_max - old_max

    self.assertLessEqual(
        growth, _RSS_GROWTH_TOLERANCE,
        msg=("The Resident Set Size (RSS) of this process grew by {} bytes, "
             "exceeding the tolerance of {} bytes."
             .format(growth, _RSS_GROWTH_TOLERANCE)))

if __name__ == "__main__":
  absltest.main()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Container classes used in control domains."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import six

_NAME_ALREADY_EXISTS = (
    "A function named {name!r} already exists in the container and "
    "`allow_overriding_keys` is False.")


class TaggedTasks(collections.Mapping):
  """Maps task names to their corresponding factory functions with tags.

  To store a function in a `TaggedTasks` container, we can use its `.add`
  decorator:

  ```python
  tasks = TaggedTasks()

  @tasks.add('easy', 'stable')
  def example_task():
    ...
    return environment

  environment_factory = tasks['example_task']

  # Or to restrict to a given tag:
  environment_factory = tasks.tagged('easy')['example_task']
  ```
  """

  def __init__(self, allow_overriding_keys=False):
    """Initializes a new `TaggedTasks` container.

    Args:
      allow_overriding_keys: Boolean, whether `add` can override existing keys
        within the container. If False (default), calling `add` multiple times
        with the same function name will result in a `ValueError`.
    """
    self._tasks = collections.OrderedDict()
    self._tags = collections.defaultdict(dict)
    self.allow_overriding_keys = allow_overriding_keys

  def add(self, *tags):
    """Decorator that adds a factory function to the container with tags.

    Args:
      *tags: Strings specifying the tags for this function.

    Returns:
      The same function.

    Raises:
      ValueError: if a function with the same name already exists within the
        container and `allow_overriding_keys` is False.
    """
    def wrap(factory_func):
      name = factory_func.__name__
      if name in self and not self.allow_overriding_keys:
        raise ValueError(_NAME_ALREADY_EXISTS.format(name=name))
      self._tasks[name] = factory_func
      for tag in tags:
        self._tags[tag][name] = factory_func
      return factory_func
    return wrap

  def tagged(self, *tags):
    """Returns a (possibly empty) dict of functions matching all the given tags.

    Args:
      *tags: Strings specifying tags to query by.

    Returns:
      A dict of `{name: function}` containing all the functions that are tagged
      by all of the strings in `tags`.
    """
    if not tags:
      return {}
    tags = set(tags)
    if not tags.issubset(six.viewkeys(self._tags)):
      return {}
    names = six.viewkeys(self._tags[tags.pop()])
    while tags:
      names &= six.viewkeys(self._tags[tags.pop()])
    return {name: self._tasks[name] for name in names}

  def tags(self):
    """Returns a list of all the tags in this container."""
    return list(self._tags.keys())

  def __getitem__(self, k):
    return self._tasks[k]

  def __iter__(self):
    return iter(self._tasks)

  def __len__(self):
    return len(self._tasks)

  def __repr__(self):
    return "{}({})".format(self.__class__.__name__, str(self._tasks))



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Base class for Walkers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import collections

from dm_control import composer
from dm_control.composer.observation import observable

from dm_env import specs
import numpy as np
import six


def _make_readonly_float64_copy(value):
  if np.isscalar(value):
    return np.float64(value)
  else:
    out = np.array(value, dtype=np.float64)
    out.flags.writeable = False
    return out


class WalkerPose(collections.namedtuple(
    'WalkerPose', ('qpos', 'xpos', 'xquat'))):
  """A named tuple representing a walker's joint and Cartesian pose."""

  __slots__ = ()

  def __new__(cls, qpos=None, xpos=(0, 0, 0), xquat=(1, 0, 0, 0)):
    """Creates a new WalkerPose.

    Args:
      qpos: The joint position for the pose, or `None` if the `qpos0` values in
        the `mjModel` should be used.
      xpos: A Cartesian displacement, for example if the walker should be lifted
        or lowered by a specific amount for this pose.
      xquat: A quaternion displacement for the root body.

    Returns:
      A new instance of `WalkerPose`.
    """
    return super(WalkerPose, cls).__new__(
        cls,
        qpos=_make_readonly_float64_copy(qpos) if qpos is not None else None,
        xpos=_make_readonly_float64_copy(xpos),
        xquat=_make_readonly_float64_copy(xquat))

  def __eq__(self, other):
    return (np.all(self.qpos == other.qpos) and
            np.all(self.xpos == other.xpos) and
            np.all(self.xquat == other.xquat))


@six.add_metaclass(abc.ABCMeta)
class Walker(composer.Robot):
  """Abstract base class for Walker robots."""

  def create_root_joints(self, attachment_frame):
    attachment_frame.add('freejoint')

  def _build_observables(self):
    return WalkerObservables(self)

  def transform_vec_to_egocentric_frame(self, physics, vec_in_world_frame):
    """Linearly transforms a world-frame vector into walker's egocentric frame.

    Note that this function does not perform an affine transformation of the
    vector. In other words, the input vector is assumed to be specified with
    respect to the same origin as this walker's egocentric frame. This function
    can also be applied to matrices whose innermost dimensions are either 2 or
    3. In this case, a matrix with the same leading dimensions is returned
    where the innermost vectors are replaced by their values computed in the
    egocentric frame.

    Args:
      physics: An `mjcf.Physics` instance.
      vec_in_world_frame: A NumPy array with last dimension of shape (2,) or
      (3,) that represents a vector quantity in the world frame.

    Returns:
      The same quantity as `vec_in_world_frame` but reexpressed in this
      entity's egocentric frame. The returned np.array has the same shape as
      np.asarray(vec_in_world_frame).

    Raises:
      ValueError: if `vec_in_world_frame` does not have shape ending with (2,)
        or (3,).
    """
    return super(Walker, self).global_vector_to_local_frame(
        physics, vec_in_world_frame)

  def transform_xmat_to_egocentric_frame(self, physics, xmat):
    """Transforms another entity's `xmat` into this walker's egocentric frame.

    This function takes another entity's (E) xmat, which is an SO(3) matrix
    from E's frame to the world frame, and turns it to a matrix that transforms
    from E's frame into this walker's egocentric frame.

    Args:
      physics: An `mjcf.Physics` instance.
      xmat: A NumPy array of shape (3, 3) or (9,) that represents another
        entity's xmat.

    Returns:
      The `xmat` reexpressed in this entity's egocentric frame. The returned
      np.array has the same shape as np.asarray(xmat).

    Raises:
      ValueError: if `xmat` does not have shape (3, 3) or (9,).
    """
    return super(Walker, self).global_xmat_to_local_frame(physics, xmat)

  @abc.abstractproperty
  def root_body(self):
    raise NotImplementedError

  @abc.abstractproperty
  def observable_joints(self):
    raise NotImplementedError

  @property
  def action_spec(self):
    minimum, maximum = zip(*[
        a.ctrlrange if a.ctrlrange is not None else (-1., 1.)
        for a in self.actuators
    ])
    return specs.BoundedArray(
        shape=(len(self.actuators),),
        dtype=np.float,
        minimum=minimum,
        maximum=maximum,
        name='\t'.join([actuator.name for actuator in self.actuators]))

  def apply_action(self, physics, action, random_state):
    """Apply action to walker's actuators."""
    del random_state
    physics.bind(self.actuators).ctrl = action


class WalkerObservables(composer.Observables):
  """Base class for Walker obserables."""

  @composer.observable
  def joints_pos(self):
    return observable.MJCFFeature('qpos', self._entity.observable_joints)

  @composer.observable
  def sensors_gyro(self):
    return observable.MJCFFeature('sensordata',
                                  self._entity.mjcf_model.sensor.gyro)

  @composer.observable
  def sensors_accelerometer(self):
    return observable.MJCFFeature('sensordata',
                                  self._entity.mjcf_model.sensor.accelerometer)

  @composer.observable
  def sensors_framequat(self):
    return observable.MJCFFeature('sensordata',
                                  self._entity.mjcf_model.sensor.framequat)

  # Semantic groupings of Walker observables.
  def _collect_from_attachments(self, attribute_name):
    out = []
    for entity in self._entity.iter_entities(exclude_self=True):
      out.extend(getattr(entity.observables, attribute_name, []))
    return out

  @property
  def proprioception(self):
    return ([self.joints_pos] +
            self._collect_from_attachments('proprioception'))

  @property
  def kinematic_sensors(self):
    return ([self.sensors_gyro,
             self.sensors_accelerometer,
             self.sensors_framequat] +
            self._collect_from_attachments('kinematic_sensors'))

  @property
  def dynamic_sensors(self):
    return self._collect_from_attachments('dynamic_sensors')




# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Constructs models for debugging_test.py.

The purpose of this file is to provide "golden" source line numbers for test
cases in debugging_test.py. When this module is loaded, it inspects its own
source code to look for lines that begin with `# !!LINE_REF`, and stores the
following line number in a dict. This allows test cases to look up the line
number by name, rather than brittly hard-coding in the line number.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os
import re

from dm_control import mjcf

SourceLine = collections.namedtuple(
    'SourceLine', ('line_number', 'text'))

LINE_REF = {}


def make_valid_model():
  # !!LINE_REF make_valid_model.mjcf_model
  mjcf_model = mjcf.RootElement()
  # !!LINE_REF make_valid_model.my_body
  my_body = mjcf_model.worldbody.add('body', name='my_body')
  my_body.add('inertial', mass=1, pos=[0, 0, 0], diaginertia=[1, 1, 1])
  # !!LINE_REF make_valid_model.my_joint
  my_joint = my_body.add('joint', name='my_joint', type='hinge')
  # !!LINE_REF make_valid_model.my_actuator
  mjcf_model.actuator.add('velocity', name='my_actuator', joint=my_joint)
  return mjcf_model


def make_broken_model():
  # !!LINE_REF make_broken_model.mjcf_model
  mjcf_model = mjcf.RootElement()
  # !!LINE_REF make_broken_model.my_body
  my_body = mjcf_model.worldbody.add('body', name='my_body')
  my_body.add('inertial', mass=1, pos=[0, 0, 0], diaginertia=[1, 1, 1])
  # !!LINE_REF make_broken_model.my_joint
  my_body.add('joint', name='my_joint', type='hinge')
  # !!LINE_REF make_broken_model.my_actuator
  mjcf_model.actuator.add('velocity', name='my_actuator', joint='invalid_joint')
  return mjcf_model


def break_valid_model(mjcf_model):
  # !!LINE_REF break_valid_model.my_actuator.joint
  mjcf_model.find('actuator', 'my_actuator').joint = 'invalid_joint'
  return mjcf_model


def _parse_line_refs():
  line_ref_pattern = re.compile(r'\s*# !!LINE_REF\s*([^\s]+)')
  filename, _ = os.path.splitext(__file__)  # __file__ can be `.pyc`.
  with open(filename + '.py') as f:
    src = f.read()
  src_lines = src.split('\n')
  for line_number, line in enumerate(src_lines):
    match = line_ref_pattern.match(line)
    if match:
      LINE_REF[match.group(1)] = SourceLine(
          line_number + 2, src_lines[line_number + 1].strip())

_parse_line_refs()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Cheetah Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from dm_control.suite import base
from dm_control.suite import common
from dm_control.utils import containers
from dm_control.utils import rewards


# How long the simulation will run, in seconds.
_DEFAULT_TIME_LIMIT = 10

# Running speed above which reward is 1.
_RUN_SPEED = 10

SUITE = containers.TaggedTasks()


def get_model_and_assets():
  """Returns a tuple containing the model XML string and a dict of assets."""
  return common.read_model('cheetah.xml'), common.ASSETS


@SUITE.add('benchmarking')
def run(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the run task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = Cheetah(random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(physics, task, time_limit=time_limit,
                             **environment_kwargs)


class Physics(mujoco.Physics):
  """Physics simulation with additional features for the Cheetah domain."""

  def speed(self):
    """Returns the horizontal speed of the Cheetah."""
    return self.named.data.sensordata['torso_subtreelinvel'][0]


class Cheetah(base.Task):
  """A `Task` to train a running Cheetah."""

  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode."""
    # The indexing below assumes that all joints have a single DOF.
    assert physics.model.nq == physics.model.njnt
    is_limited = physics.model.jnt_limited == 1
    lower, upper = physics.model.jnt_range[is_limited].T
    physics.data.qpos[is_limited] = self.random.uniform(lower, upper)

    # Stabilize the model before the actual simulation.
    for _ in range(200):
      physics.step()

    physics.data.time = 0
    self._timeout_progress = 0
    super(Cheetah, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns an observation of the state, ignoring horizontal position."""
    obs = collections.OrderedDict()
    # Ignores horizontal position to maintain translational invariance.
    obs['position'] = physics.data.qpos[1:].copy()
    obs['velocity'] = physics.velocity()
    return obs

  def get_reward(self, physics):
    """Returns a reward to the agent."""
    return rewards.tolerance(physics.speed(),
                             bounds=(_RUN_SPEED, float('inf')),
                             margin=_RUN_SPEED,
                             value_at_margin=0,
                             sigmoid='linear')



# Copyright 2020 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A (visuomotor) task consisting of reaching to targets for reward."""


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import itertools

from dm_control import composer
from dm_control.composer.observation import observable as dm_observable
import enum
import numpy as np
from six.moves import range
from six.moves import zip

DEFAULT_ALIVE_THRESHOLD = -1.0
DEFAULT_PHYSICS_TIMESTEP = 0.005
DEFAULT_CONTROL_TIMESTEP = 0.03


class TwoTouchState(enum.IntEnum):
  PRE_TOUCH = 0
  TOUCHED_ONCE = 1
  TOUCHED_TWICE = 2  # at appropriate time
  TOUCHED_TOO_SOON = 3
  NO_SECOND_TOUCH = 4


class TwoTouch(composer.Task):
  """Task with target to tap with short delay (for Rat)."""

  def __init__(self,
               walker,
               arena,
               target_builders,
               target_type_rewards,
               shuffle_target_builders=False,
               randomize_spawn_position=False,
               randomize_spawn_rotation=True,
               rotation_bias_factor=0,
               aliveness_reward=0.0,
               touch_interval=0.8,
               interval_tolerance=0.1,  # consider making a curriculum
               failure_timeout=1.2,
               reset_delay=0.,
               z_height=.14,  # 5.5" in real experiments
               target_area=(),
               physics_timestep=DEFAULT_PHYSICS_TIMESTEP,
               control_timestep=DEFAULT_CONTROL_TIMESTEP):
    self._walker = walker
    self._arena = arena
    self._walker.create_root_joints(self._arena.attach(self._walker))
    if 'CMUHumanoid' in str(type(self._walker)):
      self._lhand_body = walker.mjcf_model.find('body', 'lhand')
      self._rhand_body = walker.mjcf_model.find('body', 'rhand')
    elif 'Rat' in str(type(self._walker)):
      self._lhand_body = walker.mjcf_model.find('body', 'hand_L')
      self._rhand_body = walker.mjcf_model.find('body', 'hand_R')
    else:
      raise ValueError('Expects Rat or CMUHumanoid.')
    self._lhand_geoms = self._lhand_body.find_all('geom')
    self._rhand_geoms = self._rhand_body.find_all('geom')

    self._targets = []
    self._target_builders = target_builders
    self._target_type_rewards = tuple(target_type_rewards)
    self._shuffle_target_builders = shuffle_target_builders

    self._randomize_spawn_position = randomize_spawn_position
    self._spawn_position = [0.0, 0.0]  # x, y
    self._randomize_spawn_rotation = randomize_spawn_rotation
    self._rotation_bias_factor = rotation_bias_factor

    self._aliveness_reward = aliveness_reward
    self._discount = 1.0

    self._touch_interval = touch_interval
    self._interval_tolerance = interval_tolerance
    self._failure_timeout = failure_timeout
    self._reset_delay = reset_delay
    self._target_positions = []
    self._state_logic = TwoTouchState.PRE_TOUCH

    self._z_height = z_height
    arena_size = self._arena.size
    if target_area:
      self._target_area = target_area
    else:
      self._target_area = [1/2*arena_size[0], 1/2*arena_size[1]]
    target_x = 1.
    target_y = 1.
    self._target_positions.append((target_x, target_y, self._z_height))

    self.set_timesteps(
        physics_timestep=physics_timestep, control_timestep=control_timestep)

    self._task_observables = collections.OrderedDict()
    def task_state(physics):
      del physics
      return np.array([self._state_logic])
    self._task_observables['task_logic'] = dm_observable.Generic(task_state)

    self._walker.observables.egocentric_camera.height = 64
    self._walker.observables.egocentric_camera.width = 64

    for observable in (self._walker.observables.proprioception +
                       self._walker.observables.kinematic_sensors +
                       self._walker.observables.dynamic_sensors +
                       list(self._task_observables.values())):
      observable.enabled = True
    self._walker.observables.egocentric_camera.enabled = True

  def _get_targets(self, total_target_count, random_state):
    # Multiply total target count by the fraction for each type, rounded down.
    target_numbers = np.array([1, len(self._target_positions)-1])

    if self._shuffle_target_builders:
      random_state.shuffle(self._target_builders)

    all_targets = []
    for target_type, num in enumerate(target_numbers):
      targets = []
      if num < 1:
        break
      target_builder = self._target_builders[target_type]
      for i in range(num):
        target = target_builder(name='target_{}_{}'.format(target_type, i))
        targets.append(target)
      all_targets.append(targets)
    return all_targets

  @property
  def name(self):
    return 'two_touch'

  @property
  def task_observables(self):
    return self._task_observables

  @property
  def root_entity(self):
    return self._arena

  def _randomize_targets(self, physics, random_state=np.random):
    for ii in range(len(self._target_positions)):
      target_x = self._target_area[0]*random_state.uniform(-1., 1.)
      target_y = self._target_area[1]*random_state.uniform(-1., 1.)
      self._target_positions[ii] = (target_x, target_y, self._z_height)
    target_positions = np.copy(self._target_positions)
    random_state.shuffle(target_positions)
    all_targets = self._targets
    for pos, target in zip(target_positions, itertools.chain(*all_targets)):
      target.reset(physics)
      physics.bind(target.geom).pos = pos
    self._targets = all_targets
    self._target_rewarded_once = [
        [False] * len(targets) for targets in all_targets]
    self._target_rewarded_twice = [
        [False] * len(targets) for targets in all_targets]
    self._first_touch_time = None
    self._second_touch_time = None
    self._do_time_out = False
    self._state_logic = TwoTouchState.PRE_TOUCH

  def initialize_episode_mjcf(self, random_state):
    self._arena.regenerate(random_state)
    for target in itertools.chain(*self._targets):
      target.detach()
    target_positions = np.copy(self._target_positions)
    random_state.shuffle(target_positions)
    all_targets = self._get_targets(len(self._target_positions), random_state)
    for pos, target in zip(target_positions, itertools.chain(*all_targets)):
      self._arena.attach(target)
      target.geom.pos = pos
      target.initialize_episode_mjcf(random_state)
    self._targets = all_targets

  def _respawn_walker(self, physics, random_state):
    self._walker.reinitialize_pose(physics, random_state)

    if self._randomize_spawn_position:
      self._spawn_position = self._arena.spawn_positions[
          random_state.randint(0, len(self._arena.spawn_positions))]

    if self._randomize_spawn_rotation:
      rotation = 2*np.pi*np.random.uniform()
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]

    self._walker.shift_pose(
        physics,
        [self._spawn_position[0], self._spawn_position[1], 0.0],
        quat,
        rotate_velocity=True)

  def initialize_episode(self, physics, random_state):
    super(TwoTouch, self).initialize_episode(physics, random_state)
    self._respawn_walker(physics, random_state)
    self._state_logic = TwoTouchState.PRE_TOUCH
    self._discount = 1.0
    self._lhand_geomids = set(physics.bind(self._lhand_geoms).element_id)
    self._rhand_geomids = set(physics.bind(self._rhand_geoms).element_id)
    self._hand_geomids = self._lhand_geomids | self._rhand_geomids
    self._randomize_targets(physics)
    self._must_randomize_targets = False
    for target in itertools.chain(*self._targets):
      target._specific_collision_geom_ids = self._hand_geomids  # pylint: disable=protected-access

  def before_step(self, physics, action, random_state):
    super(TwoTouch, self).before_step(physics, action, random_state)
    if self._must_randomize_targets:
      self._randomize_targets(physics)
      self._must_randomize_targets = False

  def should_terminate_episode(self, physics):
    failure_termination = False
    if failure_termination:
      self._discount = 0.0
      return True
    else:
      return False

  def get_reward(self, physics):
    reward = self._aliveness_reward
    lhand_pos = physics.bind(self._lhand_body).xpos
    rhand_pos = physics.bind(self._rhand_body).xpos
    target_pos = physics.bind(self._targets[0][0].geom).xpos
    lhand_rew = np.exp(-3.*sum(np.abs(lhand_pos-target_pos)))
    rhand_rew = np.exp(-3.*sum(np.abs(rhand_pos-target_pos)))
    closeness_reward = np.maximum(lhand_rew, rhand_rew)
    reward += .01*closeness_reward*self._target_type_rewards[0]
    if self._state_logic == TwoTouchState.PRE_TOUCH:
      # touch the first time
      for target_type, targets in enumerate(self._targets):
        for i, target in enumerate(targets):
          if (target.activated[0] and
              not self._target_rewarded_once[target_type][i]):
            self._first_touch_time = physics.time()
            self._state_logic = TwoTouchState.TOUCHED_ONCE
            self._target_rewarded_once[target_type][i] = True
            reward += self._target_type_rewards[target_type]
    elif self._state_logic == TwoTouchState.TOUCHED_ONCE:
      for target_type, targets in enumerate(self._targets):
        for i, target in enumerate(targets):
          if (target.activated[1] and
              not self._target_rewarded_twice[target_type][i]):
            self._second_touch_time = physics.time()
            self._state_logic = TwoTouchState.TOUCHED_TWICE
            self._target_rewarded_twice[target_type][i] = True
            # check if touched too soon
            if ((self._second_touch_time - self._first_touch_time) <
                (self._touch_interval - self._interval_tolerance)):
              self._do_time_out = True
              self._state_logic = TwoTouchState.TOUCHED_TOO_SOON
            # check if touched at correct time
            elif ((self._second_touch_time - self._first_touch_time) <=
                  (self._touch_interval + self._interval_tolerance)):
              reward += self._target_type_rewards[target_type]
      # check if no second touch within time interval
      if ((physics.time() - self._first_touch_time) >
          (self._touch_interval + self._interval_tolerance)):
        self._do_time_out = True
        self._state_logic = TwoTouchState.NO_SECOND_TOUCH
        self._second_touch_time = physics.time()
    elif (self._state_logic == TwoTouchState.TOUCHED_TWICE or
          self._state_logic == TwoTouchState.TOUCHED_TOO_SOON or
          self._state_logic == TwoTouchState.NO_SECOND_TOUCH):
      # hold here due to timeout
      if self._do_time_out:
        if physics.time() > (self._second_touch_time + self._failure_timeout):
          self._do_time_out = False
      # reset/re-randomize
      elif physics.time() > (self._second_touch_time + self._reset_delay):
        self._must_randomize_targets = True
    return reward

  def get_discount(self, physics):
    del physics
    return self._discount



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for prop_initializer."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.

from absl.testing import absltest
from absl.testing import parameterized
from dm_control import composer
from dm_control import mjcf
from dm_control.composer.initializers import prop_initializer
from dm_control.composer.variation import deterministic
from dm_control.composer.variation import distributions
from dm_control.entities import props
from dm_control.rl import control
import numpy as np
from six.moves import range
from six.moves import zip


def _make_spheres(num_spheres, radius, nconmax):
  spheres = []
  arena = composer.Arena()
  arena.mjcf_model.worldbody.add('geom', type='plane', size=[1, 1, 0.1],
                                 pos=[0., 0., -2 * radius], name='ground')
  for i in range(num_spheres):
    sphere = props.Primitive(
        geom_type='sphere', size=[radius], name='sphere_{}'.format(i))
    arena.add_free_entity(sphere)
    spheres.append(sphere)
  arena.mjcf_model.size.nconmax = nconmax
  physics = mjcf.Physics.from_mjcf_model(arena.mjcf_model)
  return physics, spheres


class PropPlacerTest(parameterized.TestCase):
  """Tests for PropPlacer."""

  def assertNoContactsInvolvingEntities(self, physics, entities):
    all_colliding_geoms = set()
    for contact in physics.data.contact:
      all_colliding_geoms.add(contact.geom1)
      all_colliding_geoms.add(contact.geom2)
    for entity in entities:
      entity_geoms = physics.bind(entity.mjcf_model.find_all('geom')).element_id
      colliding_entity_geoms = all_colliding_geoms.intersection(entity_geoms)
      if colliding_entity_geoms:
        names = ', '.join(
            physics.model.id2name(i, 'geom') for i in colliding_entity_geoms)
        self.fail('Entity {} has colliding geoms: {}'
                  .format(entity.mjcf_model.model, names))

  def assertPositionsWithinBounds(self, physics, entities, lower, upper):
    for entity in entities:
      position, _ = entity.get_pose(physics)
      if np.any(position < lower) or np.any(position > upper):
        self.fail('Entity {} is out of bounds: position={}, bounds={}'
                  .format(entity.mjcf_model.model, position, (lower, upper)))

  def test_sample_non_colliding_positions(self):
    halfwidth = 0.05
    radius = halfwidth / 4.
    offset = np.array([0, 0, halfwidth + radius*1.1])
    lower = -np.full(3, halfwidth) + offset
    upper = np.full(3, halfwidth) + offset
    position_variation = distributions.Uniform(lower, upper)
    physics, spheres = _make_spheres(num_spheres=8, radius=radius, nconmax=1000)
    prop_placer = prop_initializer.PropPlacer(
        props=spheres,
        position=position_variation,
        ignore_collisions=False,
        settle_physics=False)
    prop_placer(physics, random_state=np.random.RandomState(0))
    self.assertNoContactsInvolvingEntities(physics, spheres)
    self.assertPositionsWithinBounds(physics, spheres, lower, upper)

  def test_rejection_sampling_failure(self):
    max_attempts_per_prop = 2
    fixed_position = (0, 0, 0.1)  # Guaranteed to always have collisions.
    physics, spheres = _make_spheres(num_spheres=2, radius=0.01, nconmax=1000)
    prop_placer = prop_initializer.PropPlacer(
        props=spheres,
        position=fixed_position,
        ignore_collisions=False,
        max_attempts_per_prop=max_attempts_per_prop)
    expected_message = prop_initializer._REJECTION_SAMPLING_FAILED.format(
        model_name=spheres[1].mjcf_model.model,  # Props are placed in order.
        max_attempts=max_attempts_per_prop)
    with self.assertRaisesWithLiteralMatch(RuntimeError, expected_message):
      prop_placer(physics, random_state=np.random.RandomState(0))

  def test_ignore_contacts_with_entities(self):
    physics, spheres = _make_spheres(num_spheres=2, radius=0.01, nconmax=1000)

    # Target position of both spheres (non-colliding).
    fixed_positions = [(0, 0, 0.1), (0, 0.1, 0.1)]

    # Placer that initializes both spheres to (0, 0, 0.1), ignoring contacts.
    prop_placer_init = prop_initializer.PropPlacer(
        props=spheres,
        position=fixed_positions[0],
        ignore_collisions=True,
        max_attempts_per_prop=1)

    # Sequence of placers that will move the spheres to their target positions.
    prop_placer_seq = []
    for prop, target_position in zip(spheres, fixed_positions):
      placer = prop_initializer.PropPlacer(
          props=[prop],
          position=target_position,
          ignore_collisions=False,
          max_attempts_per_prop=1)
      prop_placer_seq.append(placer)

    # We expect the first placer in the sequence to fail without
    # `ignore_contacts_with_entities` because the second sphere is already at
    # the same location.
    prop_placer_init(physics, random_state=np.random.RandomState(0))
    expected_message = prop_initializer._REJECTION_SAMPLING_FAILED.format(
        model_name=spheres[0].mjcf_model.model, max_attempts=1)
    with self.assertRaisesWithLiteralMatch(RuntimeError, expected_message):
      prop_placer_seq[0](physics, random_state=np.random.RandomState(0))

    # Placing the first sphere should succeed if we ignore contacts involving
    # the second sphere.
    prop_placer_init(physics, random_state=np.random.RandomState(0))
    prop_placer_seq[0](physics, random_state=np.random.RandomState(0),
                       ignore_contacts_with_entities=[spheres[1]])
    # Now place the second sphere with all collisions active.
    prop_placer_seq[1](physics, random_state=np.random.RandomState(0),
                       ignore_contacts_with_entities=None)
    self.assertNoContactsInvolvingEntities(physics, spheres)

  def test_exception_if_contact_buffer_always_full(self):
    max_attempts_per_prop = 2
    radius = 0.1
    num_spheres = 5
    physics, spheres = _make_spheres(num_spheres=num_spheres,
                                     radius=radius, nconmax=1)

    candidate_positions = np.multiply.outer(
        np.arange(num_spheres * max_attempts_per_prop), [radius * 2.01, 0, 0])

    # If we only place the first sphere then the others will all be overlapping
    # at the origin, so we get an error due to filling the contact buffer.
    prop_placer_failure = prop_initializer.PropPlacer(
        props=[spheres[0]],
        position=deterministic.Sequence(candidate_positions),
        ignore_collisions=False,
        max_attempts_per_prop=max_attempts_per_prop)
    with self.assertRaises(control.PhysicsError):
      prop_placer_failure(physics, random_state=np.random.RandomState(0))

    physics, spheres = _make_spheres(num_spheres=num_spheres,
                                     radius=radius, nconmax=1)

    # If we place all of the spheres then we can find a configuration where they
    # are non-colliding, so the contact buffer is not full when the initializer
    # returns.
    prop_placer = prop_initializer.PropPlacer(
        props=spheres,
        position=deterministic.Sequence(candidate_positions),
        ignore_collisions=False,
        max_attempts_per_prop=max_attempts_per_prop)
    prop_placer(physics, random_state=np.random.RandomState(0))

  def test_no_exception_if_contact_buffer_transiently_full(self):
    max_attempts_per_prop = 2
    radius = 0.1
    num_spheres = 3
    physics, spheres = _make_spheres(num_spheres=num_spheres,
                                     radius=radius, nconmax=1)
    fixed_positions = [[-radius * 1.01, 0., 0],
                       [radius * 1.01, 0., 0.]]
    for sphere, position in zip(spheres[:2], fixed_positions):
      sphere.set_pose(physics, position=position)

    candidate_positions = [
        [0., 0., 0.],  # Collides with both fixed spheres.
        [5 * radius, 0., 0.]]  # Does not collide with either sphere.

    # The first candidate position transiently fills the contact buffer.
    prop_placer = prop_initializer.PropPlacer(
        props=spheres[2:],
        position=deterministic.Sequence(candidate_positions),
        ignore_collisions=False,
        max_attempts_per_prop=max_attempts_per_prop)
    prop_placer(physics, random_state=np.random.RandomState(0))

  @parameterized.parameters([False, True])
  def test_settle_physics(self, settle_physics):
    radius = 0.1
    physics, spheres = _make_spheres(num_spheres=2, radius=radius, nconmax=1)

    # Only place the first sphere.
    prop_placer = prop_initializer.PropPlacer(
        props=spheres[:1],
        position=np.array([2.01 * radius, 0., 0.]),
        settle_physics=settle_physics)
    prop_placer(physics, random_state=np.random.RandomState(0))

    first_position, first_quaternion = spheres[0].get_pose(physics)
    del first_quaternion  # Unused.

    # If we allowed the physics to settle then the first sphere should be
    # resting on the ground, otherwise it should be at the target height.
    expected_first_z_pos = -radius if settle_physics else 0.
    self.assertAlmostEqual(first_position[2], expected_first_z_pos, places=3)

    second_position, second_quaternion = spheres[1].get_pose(physics)
    del second_quaternion  # Unused.

    # The sphere that we were not placing should not have moved.
    self.assertEqual(second_position[2], 0.)


if __name__ == '__main__':
  absltest.main()



# Copyright 2017-2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""IO functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os


def GetResource(name, mode='rb'):
  with open(name, mode=mode) as f:
    return f.read()


def GetResourceFilename(name, mode='rb'):
  del mode  # Unused.
  return name


def WalkResources(path):
  return os.walk(path)


GetResourceAsFile = open  # pylint: disable=invalid-name



# Copyright 2020 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Saves Mujoco models with relevant assets in a .zip file."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import zipfile

import six


def export_with_assets_as_zip(mjcf_model, out_dir, model_name=None):
  """Saves mjcf_model and all its assets as a .zip file in the given directory.

  Creates a .zip file named `model_name`.zip in the specified `out_dir`, and a
  directory inside of this file named `model_name`. The MJCF XML is written into
  this directory with the name `model_name`.xml, and all the assets are also
  written into this directory without changing their names.

  Args:
    mjcf_model: `mjcf.RootElement` instance to export.
    out_dir: Directory to save the .zip file. Will be created if it does not
      already exist.
    model_name: (Optional) Name of the .zip file, the name of the directory
      inside the .zip root containing the model and assets, and name of the XML
      file inside this directory. Defaults to the MJCF model name
      (`mjcf_model.model`).
  """

  if model_name is None:
    model_name = mjcf_model.model

  xml_name = model_name + '.xml'
  zip_name = model_name + '.zip'

  files_to_zip = mjcf_model.get_assets()
  files_to_zip[xml_name] = mjcf_model.to_xml_string()

  if not os.path.exists(out_dir):
    os.makedirs(out_dir)
  with zipfile.ZipFile(os.path.join(out_dir, zip_name), 'w') as zip_file:
    for filename, contents in six.iteritems(files_to_zip):
      zip_file.writestr(os.path.join(model_name, filename), contents)



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests that generated XML string is valid."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

# Internal dependencies.
from absl.testing import absltest
from dm_control.mjcf import parser
from dm_control.mujoco import wrapper

ASSETS_DIR = os.path.join(os.path.dirname(__file__), 'test_assets')
_ARENA_XML = os.path.join(ASSETS_DIR, 'arena.xml')
_LEGO_BRICK_XML = os.path.join(ASSETS_DIR, 'lego_brick.xml')
_ROBOT_XML = os.path.join(ASSETS_DIR, 'robot_arm.xml')


def validate(xml_string):
  """Validates that an XML string is a valid MJCF.

  Validation is performed by constructing Mujoco model from the string.
  The construction process contains compilation and validation phases by Mujoco
  engine, the best validation tool we have access to.

  Args:
    xml_string: XML string to validate
  """

  mjmodel = wrapper.MjModel.from_xml_string(xml_string)
  wrapper.MjData(mjmodel)


class XMLValidationTest(absltest.TestCase):

  def testXmlAttach(self):
    robot_arm = parser.from_file(_ROBOT_XML)
    arena = parser.from_file(_ARENA_XML)
    lego = parser.from_file(_LEGO_BRICK_XML)

    # validate MJCF strings before changing them
    validate(robot_arm.to_xml_string())
    validate(arena.to_xml_string())
    validate(lego.to_xml_string())

    # combine objects in complex scene
    robot_arm.find('site', 'fingertip1').attach(lego)
    arena.worldbody.attach(robot_arm)

    # validate
    validate(arena.to_xml_string())


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.composer.props.primitive."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
from dm_control import composer
from dm_control import mjcf
from dm_control.entities.props import primitive
import numpy as np


class PrimitiveTest(parameterized.TestCase):

  def _make_free_prop(self, geom_type='sphere', size=(0.1,), **kwargs):
    prop = primitive.Primitive(geom_type=geom_type, size=size, **kwargs)
    arena = composer.Arena()
    arena.add_free_entity(prop)
    physics = mjcf.Physics.from_mjcf_model(arena.mjcf_model)
    return prop, physics

  @parameterized.parameters([
      dict(geom_type='sphere', size=[0.1]),
      dict(geom_type='capsule', size=[0.1, 0.2]),
      dict(geom_type='cylinder', size=[0.1, 0.2]),
      dict(geom_type='box', size=[0.1, 0.2, 0.3]),
      dict(geom_type='ellipsoid', size=[0.1, 0.2, 0.3]),
  ])
  def test_instantiation(self, geom_type, size):
    name = 'foo'
    rgba = [1., 0., 1., 0.5]
    prop, physics = self._make_free_prop(
        geom_type=geom_type, size=size, name=name, rgba=rgba)
    # Check that the name and other kwargs are set correctly.
    self.assertEqual(prop.mjcf_model.model, name)
    np.testing.assert_array_equal(physics.bind(prop.geom).rgba, rgba)
    # Check that we can step without anything breaking.
    physics.step()

  @parameterized.parameters([
      dict(position=[0., 0., 0.]),
      dict(position=[0.1, -0.2, 0.3]),
  ])
  def test_position_observable(self, position):
    prop, physics = self._make_free_prop()
    prop.set_pose(physics, position=position)
    observation = prop.observables.position(physics)
    np.testing.assert_array_equal(position, observation)

  @parameterized.parameters([
      dict(quat=[1., 0., 0., 0.]),
      dict(quat=[0., -1., 1., 0.]),
  ])
  def test_orientation_observable(self, quat):
    prop, physics = self._make_free_prop()
    normalized_quat = np.array(quat) / np.linalg.norm(quat)
    prop.set_pose(physics, quaternion=normalized_quat)
    observation = prop.observables.orientation(physics)
    np.testing.assert_array_almost_equal(normalized_quat, observation)

  @parameterized.parameters([
      dict(velocity=[0., 0., 0.]),
      dict(velocity=[0.1, -0.2, 0.3]),
  ])
  def test_linear_velocity_observable(self, velocity):
    prop, physics = self._make_free_prop()
    prop.set_velocity(physics, velocity=velocity)
    observation = prop.observables.linear_velocity(physics)
    np.testing.assert_array_almost_equal(velocity, observation)

  @parameterized.parameters([
      dict(angular_velocity=[0., 0., 0.]),
      dict(angular_velocity=[0.1, -0.2, 0.3]),
  ])
  def test_angular_velocity_observable(self, angular_velocity):
    prop, physics = self._make_free_prop()
    prop.set_velocity(physics, angular_velocity=angular_velocity)
    observation = prop.observables.angular_velocity(physics)
    np.testing.assert_array_almost_equal(angular_velocity, observation)


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for mjbindings."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.

from absl.testing import absltest
from absl.testing import parameterized

from dm_control.mujoco.wrapper.mjbindings import constants
from dm_control.mujoco.wrapper.mjbindings import sizes


class MjbindingsTest(parameterized.TestCase):

  @parameterized.parameters(
      ('mjdata', 'xpos', ('nbody', 3)),
      ('mjmodel', 'geom_type', ('ngeom',)),
      # Fields with identifiers in mjxmacro that are resolved at compile-time.
      ('mjmodel', 'actuator_dynprm', ('nu', constants.mjNDYN)),
      # Fields with multiple named indices.
      ('mjmodel', 'key_qpos', ('nkey', 'nq')),
  )
  def testIndexDict(self, struct_name, field_name, expected_metadata):
    self.assertEqual(expected_metadata,
                     sizes.array_sizes[struct_name][field_name])


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for `dm_control.mjcf.export_with_assets`."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

# Internal dependencies.
from absl import flags
from absl.testing import absltest
from absl.testing import parameterized
from dm_control import mjcf
from dm_control.mujoco import wrapper
from dm_control.mujoco.wrapper import util
import six

FLAGS = flags.FLAGS

_ASSETS_DIR = os.path.join(os.path.dirname(__file__), 'test_assets')
_TEST_MODEL_WITH_ASSETS = os.path.join(_ASSETS_DIR, 'model_with_assets.xml')
_TEST_MODEL_WITHOUT_ASSETS = os.path.join(_ASSETS_DIR, 'lego_brick.xml')


def setUpModule():
  # Flags are not parsed when this test is invoked by `nosetests`, so we fall
  # back on using the default value for ``--test_tmpdir`.
  if not FLAGS.is_parsed():
    FLAGS.test_tmpdir = absltest.get_default_test_tmpdir()
    FLAGS.mark_as_parsed()


class ExportWithAssetsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('with_assets', _TEST_MODEL_WITH_ASSETS, 'mujoco_with_assets.xml'),
      ('without_assets', _TEST_MODEL_WITHOUT_ASSETS, 'mujoco.xml'),)
  def test_export_model(self, xml_path, out_xml_name):
    """Save processed MJCF model."""
    out_dir = self.create_tempdir().full_path
    mjcf_model = mjcf.from_path(xml_path)
    mjcf.export_with_assets(
        mjcf_model, out_dir=out_dir, out_file_name=out_xml_name)

    # Read the files in the output directory and put their contents in a dict.
    out_dir_contents = {}
    for filename in os.listdir(out_dir):
      with open(os.path.join(out_dir, filename), 'rb') as f:
        out_dir_contents[filename] = f.read()

    # Check that the output directory contains an XML file of the correct name.
    self.assertIn(out_xml_name, out_dir_contents)

    # Check that its contents match the output of `mjcf_model.to_xml_string()`.
    xml_contents = util.to_native_string(out_dir_contents.pop(out_xml_name))
    expected_xml_contents = mjcf_model.to_xml_string()
    self.assertEqual(xml_contents, expected_xml_contents)

    # Check that the other files in the directory match the contents of the
    # model's `assets` dict.
    assets = mjcf_model.get_assets()
    self.assertDictEqual(out_dir_contents, assets)

    # Check that we can construct an MjModel instance using the path to the
    # output file.
    from_exported = wrapper.MjModel.from_xml_path(
        os.path.join(out_dir, out_xml_name))

    # Check that this model is identical to one constructed directly from
    # `mjcf_model`.
    from_mjcf = wrapper.MjModel.from_xml_string(
        expected_xml_contents, assets=assets)
    self.assertEqual(from_exported.to_bytes(), from_mjcf.to_bytes())

  def test_default_model_filename(self):
    out_dir = self.create_tempdir().full_path
    mjcf_model = mjcf.from_path(_TEST_MODEL_WITH_ASSETS)
    mjcf.export_with_assets(mjcf_model, out_dir, out_file_name=None)
    expected_name = mjcf_model.model + '.xml'
    self.assertTrue(os.path.isfile(os.path.join(out_dir, expected_name)))

  def test_exceptions(self):
    out_dir = self.create_tempdir().full_path
    mjcf_model = mjcf.from_path(_TEST_MODEL_WITH_ASSETS)
    with six.assertRaisesRegex(self, ValueError, 'must end with \'.xml\''):
      mjcf.export_with_assets(mjcf_model, out_dir,
                              out_file_name='invalid_extension.png')


if __name__ == '__main__':
  absltest.main()
